# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import matplotlib.pyplot as plt
from scipy.integrate import odeint, ode
from scipy.interpolate import UnivariateSpline
import warnings
from .. import Profile, make_unit, ScalarField, VectorField, \
    TemporalVectorFields
import scipy.integrate as spinteg

NUMBERTYPES = (int, float, int, np.float16, np.float32)
ARRAYTYPES = (list, np.ndarray)


class BlasiusBL(object):
    """
    Class representing a Blasius-like boundary layer.

    Parameters
    ----------
    Uinf : number
        Flown velocity away from the wall (m/s).
    nu : number
        Kinematic viscosity (m²/s).
    rho : number
        Density (kg/m^3)
    """

    def __init__(self, Uinf, nu, rho):
        """
        Class constructor.
        """
        if not isinstance(Uinf, NUMBERTYPES):
            raise TypeError("Uinf is not a number")
        if not isinstance(nu, NUMBERTYPES):
            raise TypeError("nu is not a number")
        if not isinstance(rho, NUMBERTYPES):
            raise TypeError("rho is not a number")
        self.Uinf = Uinf
        self.nu = nu
        self.rho = rho

    def get_Rex(self, x):
        """
        Return the Reynolds number based on the distance from the beginning of
        the plate.
        """
        Re_x = self.Uinf*x/self.nu
        if Re_x > 2e5:
            print('Warning : Turbulent BL')
        return Re_x

    def get_BL_properties(self, x, allTurbulent=False, bl_perc="99"):
        """
        Return the boundary layer properties according to blasius theory.

        Parameters
        ----------
        x : number or array of number
            Position where the boundary layer thickness is computed (m)
        (can be a list).
        allTurbulent : bool, optional
            if True, the all boundary layer is considered turbulent.
        bl_perc : string in ['95', '99']
            Percentage of maximum velocity defining the BL thickness

        Returns
        -------
        delta : array of numbers
            Boundary layer thickness profile (m)
        delta2 : array of numbers
            Boundary layer momentum thickness (m)
        delta_star : array of numbers
            Boundary layer displacement thickness (m)
        H : array of numbers
            Shape factor
        Cf : array of numbers
            Friction coefficient profile
        Rex : array of numbers
            Reynolds number on the distance from the border.
        tau_w : array of numbers
            Wall shear stress (Pa)
        """
        # check
        if not isinstance(x, (NUMBERTYPES, ARRAYTYPES)):
            raise TypeError("x is not a number or a list")
        if isinstance(x, NUMBERTYPES):
            x = np.array([x])
        if not isinstance(allTurbulent, bool):
            raise TypeError("'allTurbulent' has to be a boolean")
        if bl_perc == "95":
            perc_coef_lam = 3.92
        elif bl_perc == "99":
            perc_coef_lam = 4.91
        else:
            raise ValueError()
        # Loop on c position
        delta = []
        delta2 = []
        delta_star = []
        H = np.ones(len(x))*2.59
        Cf = []
        Rex = []
        tau_w = []
        for xpos in x:
            if xpos == 0:
                delta.append(0)
                delta2.append(0)
                delta_star.append(0)
                Cf.append(0)
                Rex.append(0)
                tau_w.append(0)
            else:
                Rex.append(self.Uinf*xpos/self.nu)
                if Rex[-1] < 5e5 and not allTurbulent:
                    delta.append(xpos*perc_coef_lam/np.power(Rex[-1], 0.5))
                    delta2.append(0.664*(self.nu*xpos/self.Uinf)**.5)
                    delta_star.append(delta2[-1]*2.59)
                    Cf.append(0.664/np.power(Rex[-1], 0.5))
                    tau_w.append(Cf[-1]*1./2.*self.rho*self.Uinf**2)
                else:
                    if bl_perc == '95':
                        raise NotImplementedError()
                    delta.append(xpos*0.3806/np.power(Rex[-1], 0.2))
                    delta2.append(0)
                    delta_star.append(delta2[-1]*2.59)
                    Cf.append(0.0592/np.power(Rex[-1], 0.2))
                    tau_w.append(Cf[-1]*1./2.*self.rho*self.Uinf**2)
        delta = np.array(delta)
        delta2 = np.array(delta2)
        delta_star = np.array(delta_star)
        H = np.array(H)
        Cf = np.array(Cf)
        Rex = np.array(Rex)
        tau_w = np.array(tau_w)
        return delta, delta2, delta_star, H, Cf, Rex, tau_w

    def get_thickness_with_confinement(self, x, h, allTurbulent=False):
        """
        Return the boundary layer thickness and the friction coefficient
        according to blasius theory and adapted for use with low water levels.
        (Only valid in laminar BL)
        Fonction
        --------
        delta, Cf = BlasiusBL(allTurbulent=False)

        Parameters
        ----------
        x : number or array of number
            Position where the boundary layer thickness is computed (m)
            (can be a list).
        h : number
            Water depth (m).
        allTurbulent : bool, optional
            if True, the all boundary layer is considered turbulent.

        Returns
        -------
        delta : Profile object
            Boundary layer thickness profile (m)
        """
        if not isinstance(x, (NUMBERTYPES, ARRAYTYPES)):
            raise TypeError("x is not a number or a list")
        if isinstance(x, NUMBERTYPES):
            x = np.array([x])
        if isinstance(h, NUMBERTYPES):
            h = np.array([h])
        delta_blas = self.get_BL_properties(x, allTurbulent=allTurbulent)[0]
        delta_perso = delta_blas*(1 - 0.26547*delta_blas/h)
        # returning
        delta = Profile(x, delta_perso, unit_x=make_unit('m'),
                        unit_y=make_unit('m'))
        return delta

    def get_profile(self, x, y=None, allTurbulent=False):
        """
        Return a Blasius-like (laminar) profile at the given position.

        Parameters
        ----------
        x : number
            Position of the profile along x axis
        y : array of numbers
            Point along y where to compute the profile (if not specified,
            200 homogeneously placed points are used)
        allTurbulent : bool, optional
            if True, the boundary layer is considered turbulent.

        Returns
        -------
        prof : Profile Object
            Wanted Blasius-like profile.
        """
        # check
        if not isinstance(x, NUMBERTYPES):
            raise TypeError()
        # Not turbulent case
        if not allTurbulent:
            def f_deriv(F, theta):
                """
                y' = dy/dx
                y'' = dy'/dx
                dy''/dx = -1/2*y*dy'/dx
                """
                return [F[1], F[2], -1./2.*F[0]*F[2]]
            # profile initial values
            f0 = [0, 0, 0.332]
            # y values
            if y is None:
                theta = np.linspace(0, 10, 200)
                y = theta*np.sqrt(x)*np.sqrt(self.nu/self.Uinf)
            else:
                theta = y/(np.sqrt(x)*np.sqrt(self.nu/self.Uinf))
            # solving with scipy ode solver
            sol = odeint(f_deriv, f0, theta)
            # getting adimensionnale velocity
            u_over_U = sol[:, 1]
            # getting dimensionnal values
            u = u_over_U*self.Uinf
        # Turbulent case
        else:
            delta, _, _ = self.get_thickness(x, allTurbulent=True)
            if y is None:
                theta = np.linspace(0, 10, 200)
                y = theta*delta.y[0]
            else:
                theta = y/delta.y[0]
            u_over_U = np.power(theta, 1./7.)
            u_over_U[theta > 1] = 1.
            u = u_over_U*self.Uinf
        return Profile(y, u, unit_x=make_unit('m'), unit_y=make_unit('m/s'))

    def get_profile_with_confinement(self, x, h, y=None, allTurbulent=False):
        """
        Return a Blasius-like (laminar) profile at the given position, ajusted
        for confined BL.

        Parameters
        ----------
        x : number
            Position of the profile along x axis
        h : number
            Pater level.
        y : array of numbers
            Point along y where to compute the profile (if not specified,
            200 homogeneously placed points are used)
        allTurbulent : bool, optional
            if True, the boundary layer is considered turbulent.

        Returns
        -------
        prof : Profile Object
            Wanted Blasius-like profile.
        """
        # check
        if not isinstance(x, NUMBERTYPES):
            raise TypeError()
        if not isinstance(h, NUMBERTYPES):
            raise TypeError()
        # get
        delta = self.get_thickness_with_confinement(x, h,
                                                    allTurbulent=allTurbulent)
        eq_x = self.get_x_from_delta(delta.y[0])
        # return
        return self.get_profile(eq_x, y=y, allTurbulent=allTurbulent)

    def get_x_from_delta(self, delta, allTurbulent=False):
        """
        Return a the x value that give the wanted delta.
        """
        # getting the laminar value of x
        xpos = (delta**2*self.Uinf)/(4.92**2*self.nu)
        # checking if turbulent
        Re_x = self.get_Rex(xpos)
        if Re_x > 5e5 or allTurbulent:
            xpos = ((delta*self.Uinf**(.2))/(0.3806*self.nu**(.2)))**(1./.8)
        return xpos


class FalknerSkanBL(object):

    def __init__(self, nu, m, c0, L):
        """
        Represent a Boundary layer with a pressure gradient, according to
        Falkner-Skan.

        Parameters
        ----------
        nu : number
            Viscosity
        m, c0 , L: integer and numbers
            Pressure gradient parameters. Such as U_e = c0*(x/L)^m.
        """
        self.nu = nu
        self.m = m
        self.c0 = c0
        self.L = L
        self.beta = (2*self.m)/(self.m + 1)

    def get_u_e(self, x):
        """
        Return the external velocity at position 'x'.
        """
        return self.c0*(x/self.L)**self.m

    def get_mu(self, x, y):
        """
        return the \mu parameter at the position '(x, y)'.
        """
        return y*(self.c0*(self.m + 1)/(2*self.nu*self.L))**.5 * \
            (x/self.L)**((self.m - 1)/2.)

    def get_y(self, x, mu):
        """
        Return the 'y' vaue associated with \mu parameter.
        """
        return mu*(self.nu*x/self.get_u_e(x))**.5
        return mu/((self.c0*(self.m + 1)/(2*self.nu*self.L))**.5 *
                   (x/self.L)**((self.m - 1)/2.))

    def get_f_function(self, relerr=1e-5, max_it=1000, verbose=False):
        """
        Return the 'f' function appearing in the Falkner-skan equation, and its
        derivatives.

        Notes
        -----
        * Falkner-Skan equation :

        .. math::
            f''' + ff'' + \\beta \\left( 1 - f'^2 \\right) = 0

            f(0) = f'(0) = 0 \\: \\text{ and } \\: f'(\\infty) = 1

        * Associated constants :

        .. math::
            \\beta = \\frac{2m}{m + 1}

            -0.0905 \le m \le 2

        * Remark on numerical resolution :

            The Falkner-Skan equation is a ODE system with BVP
            (boundary value problem). Classical ODE algorithm such as
            Runge-Kutta or Vode cannot take care of the :math:`f'(\\infty)=0`.
            This ODE system is so solved with a shooting method.
        """
        import warnings
        warnings.filterwarnings('error')
        # Set mu values
        mu = np.linspace(0, 10, 100)

        # Implement Falkner-Skan equation
        def deriv(Y, t):
            Y_prime = [0, 0, 0]
            Y_prime[0] = Y[1]
            Y_prime[1] = Y[2]
            Y_prime[2] = - (self.m + 1)/2.*Y[0]*Y[2] - self.m*(1. - Y[1]**2)
            return Y_prime

        def deriv2(t, Y):
            return deriv(Y, t)

        # Create ODE system for resolution
        integrator = "lsoda"
        ode_maxsteps = 10000
        ODE = ode(deriv2)
        ODE.set_integrator(integrator, nsteps=ode_maxsteps)

        # Prepare shooting method to solve BVP (boundary value problem)
        aimed_Y1p = 1.
        first_Y3_guess = 0.33 + 0.77*np.log(2.207*(self.m + 0.45696))
        interval_Y3 = [first_Y3_guess - .33, first_Y3_guess + .33]
        # first guesses for f'''(0)
        tmp_Y3s = np.linspace(*interval_Y3, num=100)
        tmp_Y1ps = []
        tmp_Y2ps = []
        for tmpY3 in tmp_Y3s:
            ODE.set_initial_value([0, 0, tmpY3])
            try:
                prof = ODE.integrate(mu[-1])
            except RuntimeWarning:
                tmp_Y1ps.append(np.nan)
                tmp_Y2ps.append(np.nan)
            else:
                tmp_Y1ps.append(prof[1])
                tmp_Y2ps.append(prof[2])
        tmp_Y1ps = np.array(tmp_Y1ps)
        tmp_Y2ps = np.array(tmp_Y2ps)[~np.isnan(tmp_Y1ps)]
        tmp_Y3s = tmp_Y3s[~np.isnan(tmp_Y1ps)]
        tmp_Y1ps = tmp_Y1ps[~np.isnan(tmp_Y1ps)]
        # check if we get a cool interval with apropriate derivates
        resid = tmp_Y1ps - aimed_Y1p
        ind_min = np.argsort(np.abs(tmp_Y1ps - aimed_Y1p))
        if np.min(resid)*np.max(resid) < 0:
            tmp_neg = resid[:-1]*resid[1::]
            ind1 = np.where(tmp_neg < 0)[0]
            ind2 = np.where(tmp_Y2ps[:-1]*tmp_Y2ps[1:] < 0)[0]
            ind = [ind1[i]
                   for i in np.arange(len(ind1))
                   if ind1[i] == ind2[i]][0]
            interval_Y1p = [tmp_Y1ps[ind], tmp_Y1ps[ind + 1]]
            interval_Y3 = [tmp_Y3s[ind], tmp_Y3s[ind + 1]]
        # check if we get a lesser cool interval
        elif ind_min[0] in [0, len(tmp_Y1ps) - 1]:
            raise Exception("Guessing interval for f''(0) need to be enlarged"
                            "\n tmp_Y1ps={}".format(tmp_Y1ps))
        else:
            interval_Y1p = [tmp_Y1ps[ind_min[0]], tmp_Y1ps[ind_min[1]]]
            interval_Y3 = [tmp_Y3s[ind_min[0]], tmp_Y3s[ind_min[1]]]

        # Shooting method to solve iteratively boundary conditions
        nmb_it = 0
        if verbose:
            print("+++ Shooting method iteration +++")
        mid_Y3 = 1e30
        while True:
            new_mid_Y3 = ((interval_Y3[0]*np.abs(1 - interval_Y1p[1]) +
                           interval_Y3[1]*np.abs(1 - interval_Y1p[0])) /
                          (np.abs(1 - interval_Y1p[1]) +
                           np.abs(1 - interval_Y1p[0])))
            if np.abs(new_mid_Y3 - mid_Y3) < 1e-10:
                if verbose:
                    print("+++    Can't do better than that "
                          "(stuck in local minimum ?)")
                break
            mid_Y3 = new_mid_Y3
            if verbose:
                print(("+++    ({})     new tested Y3 = {:.4f}"
                       .format(nmb_it, mid_Y3)))
            # compute new Y1p at middle
            ODE.set_initial_value([0, 0, mid_Y3])
            prof = ODE.integrate(mu[-1])
            new_Y1p = prof[1]
            # replace interval with new values
            if (interval_Y1p[0] - aimed_Y1p)*(new_Y1p - aimed_Y1p) < 0:
                interval_Y1p[1] = new_Y1p
                interval_Y3[1] = mid_Y3
            else:
                interval_Y1p[0] = new_Y1p
                interval_Y3[0] = mid_Y3
            err = np.abs((new_Y1p - aimed_Y1p)/np.mean([new_Y1p, aimed_Y1p]))
            if verbose:
                print(("+++             err = {:.4f}".format(err)))
            # stopping test
            if nmb_it > max_it:
                if verbose:
                    print("+++    Maximum number of iteration reached")
                break
            if err < relerr:
                if verbose:
                    print(("+++    Converged in {} iterations !"
                           .format(nmb_it)))
                    break
            # incr
            nmb_it += 1
        Y3 = mid_Y3

        # Solve ODE with the new set of boundary conditions : f(0)=f'(0)= 0 and
        # f''(0) = Y3
        sol = odeint(deriv, y0=[0, 0, Y3], t=mu)

        # Return
        return mu, sol

    def get_profile(self, x, relerr=1e-5, max_it=1000, verbose=False):
        """
        return the velocity profile at 'x' position.
        """
        # solve ODE system using shooting method and Vode
        mu, sol = self.get_f_function(relerr=relerr, max_it=max_it,
                                      verbose=verbose)
        # get y values
        y = self.get_y(x, mu)

        # check solution
        dmu = mu[1] - mu[0]
        F0 = sol[:, 0]
        F1 = sol[:, 1]
        F2 = sol[:, 2]
        F3 = np.gradient(F2, dmu)
        res = np.mean(F3 + (self.m + 1)/2*F0*F2 + self.m*(1 - F1**2))
        print("+++ Equation +++")
        print(("+++     should be zero : {:.4f}".format(res)))
        print("+++ Boundary conditions +++")
        print(("+++     f(0)={:.3f} (should be 0)".format(sol[0, 0])))
        print(("+++     f'(0)={:.3f} (should be 0)".format(sol[0, 1])))
        print(("+++     f'(inf)={:.3f} (should be 1)".format(sol[-1, 1])))
        vx = sol[:, 1]*self.get_u_e(x)
        return y, vx


class ThwaitesBL(object):

    def __init__(self, u_e, nu=1.e-6):
        """
        Represent a The solution of the boundary layer momentum equation
        using Thwaites solution.

        Parameters
        ----------
        u_e : function
            External velocity, should take a distance in 'm' as argument
            and returning a velocity in 'm/s'.
        nu : number
            Viscosity (default to water, at 1e-6)
        """
        self.nu = nu
        self.u_e = u_e
        # classical empirical coefficients
        self.emp_coef_1 = 0.441
        self.emp_coef_2 = 5.68
#        # adapted empirical coefficients
#        self.emp_coef_1 = 0.19199
#        self.emp_coef_2 = 10.6612

    def u_e_pow(self, x):
        """
        Function u_e(x)**4.68
        (used int numerical resolution)
        """
        return self.u_e(x)**(self.emp_coef_2 - 1)

    def get_momentum_thikness(self, x):
        """
        Return the evolution of the boundary layer momentum thikness.

        Parameters
        ----------
        x : number or array of numbers
            Position along the flat plan until where we want to solve the
            momentum equation.
            (has no influence on the resoluion accuracy)

        Returns
        -------
        delta2 : position and value of momentum thicknesses
        """
        if isinstance(x, NUMBERTYPES):
            res, eps = spinteg.quad(func=self.u_e_pow, a=0, b=x)
            denom = self.u_e(x)**self.emp_coef_2
            if denom == 0:
                delta2 = None
            else:
                delta2 = (self.emp_coef_1*self.nu/denom*res)**.5
        elif isinstance(x, ARRAYTYPES):
            x = np.array(x, dtype=float)
            delta2 = []
            for xi in x:
                delta2.append(self.get_momentum_thikness(xi))
            delta2 = np.array(delta2, dtype=float)
        else:
            raise TypeError()
        return np.array(delta2)

    def get_BL_properties(self, x):
        """
        Return the boundary layer properties for given values of x

        Parameters
        ----------
        x : number or array of numbers
            Position along the flat plan until where we want to solve the
            momentum equation.
            (has no influence on the resoluion accuracy)

        Returns
        -------
        lambda : array of numbers
            Dimensionless pressure gradient.
            (According to Kays and Crawford, a value of -0.082 is caracteristic
            of the separation phenomenon.)
        delta2 : array of numbers
            Boundary layer momentum thickness
        delta_star : array of numbers
            Boundary layer displacement thickness
        H : array of numbers
            H factor
        """
        # get u_e and grad u_e
        u_e = np.array([self.u_e(xi) for xi in x])
        d_u_e = np.gradient(u_e)
        # get delta2
        delta2 = self.get_momentum_thikness(x)
        # get lambda
        lambd = delta2**2/self.nu*d_u_e
        # get H factor
        H = []
        for lambdi in lambd:
            if lambdi >= 0 and lambdi < 0.1:
                H.append(2.61 - 3.75*lambdi + 5.24*lambdi**2)
            elif lambdi < 0 and lambdi > -0.1:
                H.append(2.088 + 0.0731/(lambdi + 0.14))
            else:
                H.append(np.nan)
        H = np.array(H, dtype=float)
        # get delta_star
        delta_star = H*delta2
        return lambd, delta2, delta_star, H


class WallLaw(object):
    """
    Class representing a law of the wall profile.
    By default, the used liquid is water.

    Parameters
    ----------
    h : number
        Water depth (m)
    tau : number
        The wall shear stress (Pa)
    visc_c : number, optional
        Kinematic viscosity (m²/s)
    rho : number, optional
        liquid density (kg/m^3)
    """

    def __init__(self, h, tau, delta, visc_c=1e-6, rho=1000):
        """
        Class constructor.
        """
        if not isinstance(h, NUMBERTYPES):
            raise TypeError("'h' has to be a number")
        if h <= 0:
            raise ValueError("'h' has to be positive")
        if not isinstance(tau, NUMBERTYPES):
            raise TypeError("'tau' has to be a number")
        if not isinstance(delta, NUMBERTYPES):
            raise TypeError("'delta' has to be a number")
        if tau < 0:
            raise ValueError("'tau' has to be a positive number")
        self.k = 0.4
        self.A = 5.5
        self.rho = rho
        self.tau = tau
        self.delta = delta
        self.Utau = np.sqrt(self.tau/self.rho)
        self.visc_c = visc_c
        self.visc_d = self.visc_c*self.rho
        self.h = h

    def display(self, dy, **plotArgs):
        """
        Display the velocity profile.

        Parameters
        ----------
        dy : number
            Resolution along the water depth (m).

        Returns
        -------
        fig : figure reference
            Reference to the displayed figure.
        """

        if not isinstance(dy, NUMBERTYPES):
            raise TypeError("'dy' has to be a number")
        if dy < 0:
            raise ValueError("'dy' has to be a positive number")
        if dy > self.h:
            raise ValueError("'dy' has to be smaller than the water depth")
        y = np.arange(0, self.h, dy)
        prof = self.get_profile(y)
        Umoy, _ = prof.get_integral()
        fig = prof.display(reverse=True, label=("tau={0:.4f} : "
                           "Umoyen = {1:.4f}").format(self.tau, Umoy))
        y5 = 5.*self.visc_c*np.sqrt(self.rho/self.tau)
        y30 = 30.*self.visc_c*np.sqrt(self.rho/self.tau)
        mini = prof.get_min()
        maxi = prof.get_max()
        plt.plot([mini, maxi], [y5, y5], 'r--')
        plt.plot([mini, maxi], [y30, y30], 'r--')
        plt.xlabel("Velocity (m/s)")
        plt.ylabel("y (m)")
        plt.title("velocity profile according to the log-defect law")
        return fig

    def get_profile(self, y):
        """
        Return a log-defect profile, according to the given parameters.

        Parameters
        ----------
        y : array
            Value of y in which calculate the profile (m).

        Returns
        -------
        prof : Profile object
            the profile for values of 'y'
        """
        if not isinstance(y, ARRAYTYPES):
            raise TypeError("'y' has to be an array")
        if any(y < 0):
            raise ValueError("'y' has to be an array of positive number")
        if any(y > self.h):
            raise ValueError("'y' has to be smaller than the water depth")
        y = np.array(y)
        yplus = y*self.Utau/self.visc_c
        ylimscv = 11.63
        ylimlog = 450.
        Ufin = []
        for i, yp in enumerate(yplus):
            if yp < 0:
                Utmp = 0
            elif yp <= ylimscv:
                Utmp = self._scv(yp)*self.Utau
            elif yp <= ylimlog:
                Utmp = self._inner_turb(yp)*self.Utau
            elif y[i] <= self.delta:
                Utmp = self._outer_turb(yp)*self.Utau
            else:
                Utmp = self._undisturbed(yp)*self.Utau
            Ufin.append(Utmp)
        Ufin = Profile(y, Ufin, mask=False, unit_x=make_unit("m"),
                       unit_y=make_unit("m/s"))
        return Ufin

    def integral(self, x, y):
        return np.trapz(y, x)

    def _scv(self, yp):
        """
        Calculate u/Utau in the scv.
        """
        return yp

    def _inner_turb(self, yp):
        """
        Calculate u/Utau in the inner part of the bl.
        """
        return 1./self.k*np.log(yp) + 5.1

    def _outer_turb(self, yp):
        """
        Calculate u/Utau in the outer part of the bl.
        """
        return self._inner_turb(yp)

    def _undisturbed(self, yp):
        """
        Calculate u/Utau outside of the bl
        """
        return self._outer_turb(self.delta*self.Utau/self.visc_c)


class WakeLaw(WallLaw):
    """
    Class representing a law of the wake profile using Coles theory.
    By default, the used liquid is water.

    Parameters
    ----------
    h : number
        Water depth (m)
    tau : number
        The wall shear stress (Pa)
    delta : number
        The boundary layer thickness (m)
    Cc : number, optional
        The Coles parameters (n.u) (0.45 by default)
    visc_c : number, optional
        Kinematic viscosity (m²/s) (defaul = 1e-6)
    rho : number, optional
        liquid density (kg/m^3) (default = 1000)
    """

    def _inner_turb(self, yp):
        """
        Calculate u/Utau in the turbulent part of the bl.
        """
        return self._wake_law(yp)

    def _outer_turb(self, yp):
        """
        Calculate u/Utau outside of the bl.
        """
        return self._wake_law(yp)

    def _wake_law(self, yp):
        """
        Calculate the defect composante of the law.
        Take yp.
        Return u/Utau.
        """
        y = yp/self.Utau*self.visc_c
        Cc = 0.55
        return (1./self.k*np.log(self.Utau*y/self.visc_c) + 5.1 +
                2.*Cc/self.k*np.sin(np.pi*y/(2.*self.delta))**2)


def get_bl_thickness(obj, direction=1, perc=0.95):
    """
    Return a boundary layer thickness if 'obj' is a Profile.
    Return a profile of boundary layer thicknesses if 'obj' is a ScalarField.
    WARNING : the wall must be at x=0.

    Parameters
    ----------
    obj : Profile or ScalarField object
        Vx field.
    direction : integer, optional
        If 'obj' is a ScalarField, determine the swept axis
        (1 for x and 2 for y).
    perc : float, optionnal
        Percentage used in the bl calculation (95% per default).

    Returns
    -------
    BLT : float or profile
        Boundary layer thickness, in axe x unit.
    """
    if isinstance(obj, Profile):
        maxi = obj.max
        if maxi is None:
            return 0
        value = obj.get_interpolated_values(y=maxi*perc)
        return value[0]
    elif isinstance(obj, ScalarField):
        if direction == 1:
            axe = obj.axe_x
        else:
            axe = obj.axe_y
        profiles = [obj.get_profile(direction, x) for x in axe]
        values = [get_bl_thickness(prof, perc=perc) for prof in profiles]
        return Profile(axe, values, unit_x=obj.unit_x, unit_y=obj.unit_y)
    else:
        raise TypeError("Can't compute (yet ?) BL thickness on this kind of"
                        " data : {}".format(type(obj)))


def get_clauser_thickness(obj, direction=1, rho=1000, nu=1e-6, tau=None):
    """
    Return the profile Clauser's thickness defined in 'Clauser (1956)'.
    (Delta_star = integrale_0_h (u_top - u)/u_star dy)

    Parameters
    ----------
    obj : Profile or ScalarField object
    direction : integer, optional
        If 'obj' is a ScalarField, determine the swept axis
        (1 for x and 2 for y).
    rho : number, optional
        Density of the fluid (default fo water : 1000 kg/m^3)
    nu : number, optional
        Kinematic viscosity for the fluid (default for water : 1e-6 m^2/s)
    tau : number, optional
        Wall shear stress, if not specified, 'get_shear_stress' is used to
        compute it.

    Returns
    -------
    Delta_star : float or Profile
        Boundary layer Clauser thickness, in axe x unit.
    """
    # if obj is a profile, getting Delta_star
    if isinstance(obj, Profile):
        # getting u_star
        if tau is None:
            tau = get_shear_stress(obj, direction=direction, nu=nu, rho=rho)
            tau.change_unit('y', 'kg/m/s**2')
            tau = tau.y[0]
        u_star = np.sqrt(tau/rho)
        # getting v_top
        v_top = obj.y[-1]
        Delta_star = get_displ_thickness(obj)*v_top/u_star
        return Delta_star
    # if obj is a scalarField
    elif isinstance(obj, ScalarField):
        if direction == 1:
            axe = obj.axe_x
        else:
            axe = obj.axe_y
        profiles = [obj.get_profile(direction, x) for x in axe]
        values = [get_clauser_thickness(prof, direction=direction, rho=rho,
                                        nu=nu)
                  for prof, _ in profiles]
        return Profile(axe, values, unit_x=obj.unit_x, unit_y=obj.unit_y)


def get_displ_thickness(obj, direction=1):
    """
    Return a displacement thickness if 'obj' is a Profile.
    Return a profile of displacement thicknesses if 'obj' is a Scalarfield.
    WARNING : the wall must be at x=0.

    Parameters
    ----------
    obj : Profile or ScalarField object
    direction : integer, optional
        If 'obj' is a ScalarField, determine the swept axis
        (1 for x and 2 for y).

    Returns
    -------
    delta : float or Profile
        Boundary layer displacement thickness, in axe x unit.
    """
    if isinstance(obj, Profile):
        bl_perc = 0.95
        # cut the profile in order to only keep the BL
        bl_thick = get_bl_thickness(obj, perc=bl_perc)
        if bl_thick == 0:
            return 0
        obj = obj.crop([0, bl_thick])
        # removing negative and masked points
        if isinstance(obj.y, np.ma.MaskedArray):
            mask = np.logical_and(obj.y.mask, obj.x < 0)
            obj.x = obj.x[~mask]
            obj.y = obj.y._data[~mask]
        # if there is no more value in the profile (all masked)
        if len(obj.x) == 0:
            return 0
        # adding a x(0) value if necessary
        if obj.x[0] != 0:
            pos_x = np.append([0], obj.x)
            pos_y = np.append([0], obj.y)
        else:
            pos_x = obj.x
            pos_y = obj.y
        # computing bl displacement thickness
        fonct = 1 - pos_y/np.max(pos_y)
        delta = np.trapz(fonct, pos_x)
        return delta
    elif isinstance(obj, ScalarField):
        if direction == 1:
            axe = obj.axe_x
        else:
            axe = obj.axe_y
        profiles = [obj.get_profile(direction, x) for x in axe]
        values = [get_displ_thickness(prof) for prof in profiles]
        return Profile(axe, values, unit_x=obj.unit_x, unit_y=obj.unit_y)
    else:
        raise TypeError("Can't compute (yet ?) BL displacement thickness on"
                        "this kind of data : {}".format(type(obj)))


def get_momentum_thickness(obj, direction=1):
    """
    Return a momentum thickness if 'obj' is a Profile.
    Return a profile of momentum thicknesses if 'obj' is a Scalarfield.
    WARNING : the wall must be at x=0.

    Parameters
    ----------
    obj : Profile or ScalarField object
    direction : integer, optional
        If 'obj' is a ScalarField, determine the swept axis
        (1 for x and 2 for y).

    Returns
    -------
    delta : float or Profile
        Boundary layer momentum thickness, in axe x unit.
    """
    if isinstance(obj, Profile):
        bl_perc = 0.95
        # cut the profile in order to only keep the BL
        bl_thick = get_bl_thickness(obj, perc=bl_perc)
        if bl_thick == 0:
            return 0
        obj = obj.crop([0, bl_thick])
        # removing negative and masked points
        if isinstance(obj.y, np.ma.MaskedArray):
            mask = np.logical_and(obj.y.mask, obj.x < 0)
            obj.x = obj.x[~mask]
            obj.y = obj.y._data[~mask]
        # if there is no more profile (all masked)
        if len(obj.x) == 0:
            return 0
        # adding a x(0) value
        if obj.x[0] != 0:
            pos_x = np.append([0], obj.x)
            pos_y = np.append([0], obj.y)
        else:
            pos_x = obj.x
            pos_y = obj.y
        # computing bl momentum thickness
        fonct = pos_y/np.max(pos_y)*(1 - pos_y/np.max(pos_y))
        delta = np.trapz(fonct, pos_x)
        return delta
    elif isinstance(obj, ScalarField):
        if direction == 1:
            axe = obj.axe_x
        else:
            axe = obj.axe_y
        profiles = [obj.get_profile(direction, x) for x in axe]
        values = [get_momentum_thickness(prof) for prof in profiles]
        return Profile(axe, values, unit_x=obj.unit_x, unit_y=obj.unit_y)
    else:
        raise TypeError("Can't compute (yet ?) BL momentum thickness on"
                        "this kind of data")


def get_shape_factor(obj, direction=1):
    """
    Return a shape factor if 'obj' is a Profile.
    Return a profile of shape factors if 'obj' is a Scalarfield.
    WARNING : the wall must be at x=0.

    Parameters
    ----------
    obj : Profile or ScalarField object
    direction : integer, optional
        If 'obj' is a ScalarField, determine the swept axis
        (1 for x and 2 for y).

    Returns
    -------
    delta : float or Profile
        Boundary layer shape factor, in axe x unit.
    """
    displ = get_displ_thickness(obj, direction)
    mom = get_momentum_thickness(obj, direction)
    shape_factor = displ/mom
    if isinstance(shape_factor, Profile):
        shape_factor.mask = np.logical_or(shape_factor.mask, mom.y <= 0)
    return shape_factor


def get_shear_stress(obj, direction=1, method='simple',
                     respace=False, tau_w_guess=1e-6, rho=1000., nu=1.e-6):
    """
    Return the wall shear stress.
    If velocities values are missing near the wall, an extrapolation
    (bad accuracy) is used.
    Warning : the wall must be at x=0

    Parameters
    ----------
    obj : Profile or ScalarField object
        .
    viscosity : number, optional
        Dynamic viscosity (default to water : 1e-3)
    direction : integer, optional
        If 'obj' is a ScalarField, determine the swept axis
        (1 for x and 2 for y).
    method : string, optional
        'simple' (default) : use simple gradient computation
        'wall_law_lin' : use the linear part of the 'law of the wall' model
        (need some points in the viscous sublayer)
        'wall_law_log' : use the log part of the 'law of the wall' model
        (only valid in the log layer)
    respace : bool, optional
        Use linear interpolation to create an evenly spaced profile.
    tau_w_guess : number, optional
        For 'Wall_law_log' method, initial guess for tau_w resolution.
    rho : number, optional
        Density of the fluid (default fo water : 1000 kg/m^3)
    nu : number, optional
        Kinematic viscosity for the fluid (default for water : 1e-6 m^2/s)
    """
    unit_visc = make_unit('m^2/s')
    unit_rho = make_unit('kg/m^3')
    # check parameters
    if not isinstance(nu, NUMBERTYPES):
        raise TypeError()
    if nu <= 0:
        raise ValueError()
    if direction not in [1, 2]:
        raise ValueError()
    # if obj is a profile
    if isinstance(obj, Profile):
        if method == 'simple':
            # respace if asked
            if respace:
                obj = obj.evenly_space('linear')
            # compute gradients and return shear stress
            tmp_prof = obj.get_gradient()*nu*rho*unit_visc*unit_rho
            return tmp_prof
        elif method == 'wall_law_lin':
            new_x = obj.x[obj.x > 0]
            new_y = obj.y[obj.x > 0]/new_x*nu*rho
            new_unit_y = obj.unit_y/obj.unit_x*unit_visc*unit_rho
            mask = obj.mask[obj.x > 0]
            return Profile(x=new_x, y=new_y, mask=mask, unit_x=obj.unit_x,
                           unit_y=new_unit_y, name=obj.name)
        elif method == 'wall_law_log':
            # getting data
            import scipy.optimize as spopt
            x = obj.x[obj.x > 0]
            y = obj.y[obj.x > 0]
            mask = obj.mask[obj.x > 0]

            # log law of the wall
            def func(u_star, U, rho, y, nu):
                u_star = u_star[0]
                k = 0.41
                C = 5.1
                # compute residual
                if u_star < 0:
                    res = -(U/u_star - 1./k*np.log(np.abs(y*u_star/nu)) -
                            C)
                elif u_star == 0:
                    res = U/u_star - 1./k*np.log(-1e5) - C
                else:
                    res = U/u_star - 1./k*np.log(y*u_star/nu) - C
                return res
            # solving
            u_stars = np.zeros(len(x))
            for i in np.arange(len(u_stars)):
                u_stars[i] = spopt.fsolve(func, (1e-6,),
                                          (y[i], rho, x[i], nu))
            tau_w = rho*np.array(u_stars)**2
            unit_tau = obj.unit_y/obj.unit_x*unit_visc*unit_rho
            # returning
            return Profile(x=x, y=tau_w, mask=mask, unit_x=obj.unit_x,
                           unit_y=unit_tau, name=obj.name)
        else:
            raise ValueError()
    elif isinstance(obj, ScalarField):
        # get axis
        if direction == 1:
            axe = obj.axe_x
        else:
            axe = obj.axe_y
        # get profiles
        profiles = [obj.get_profile(direction, pos_ind, ind=True)
                    for pos_ind in range(len(axe))]
        # get shear stresses
        values = np.zeros(obj.shape)
        for i in range(len(profiles)):
            if direction == 1:
                values[i, :] = get_shear_stress(profiles[i], direction=1,
                                                method=method, respace=False,
                                                tau_w_guess=tau_w_guess,
                                                rho=rho, nu=nu).y
            else:
                values[:, i] = get_shear_stress(profiles[i], direction=1,
                                                method=method, respace=False,
                                                tau_w_guess=tau_w_guess,
                                                rho=rho, nu=nu).y
        # returning
        mask = np.isnan(values)
        unit_values = profiles[0].unit_y
        tau_w = ScalarField()
        tau_w.import_from_arrays(obj.axe_x, obj.axe_y, values, mask=mask,
                                 unit_x=obj.unit_x, unit_y=obj.unit_y,
                                 unit_values=unit_values)
        return tau_w
    else:
        raise TypeError()


def get_separation_position(obj, wall_direction, wall_position,
                            interval=None, nmb_lines=4):
    """
    Compute and return the separation points position.
    Separation points position is computed by searching zero streamwise
    velocities on surrounding field lines and by extrapolating at
    the wanted 'wall_position'.
    If specified, 'interval' must include separation points on the 4 nearest
    field line.
    If multiples changments of streamwise velocity are found, return the mean
    positions of those points.

    Parameters
    ----------
    obj : ScalarField, VectorField, VectorField or TemporalVelocityField
        If 'VectorField' or 'VectorField', wall_direction is used to
        determine the interesting velocity component.
    wall_direction : integer
        1 for a wall at a given value of x,
        2 for a wall at a given value of y.
    wall_position : number
        Position of the wall.
    interval : 2x1 array of numbers, optional
        Optional interval in which search for the detachment points.
    nmb_lines : int
        Number of lines to take into account to make the extrapolation.
        (default is 4)
    """
    # checking parameters coherence
    if not isinstance(obj, (ScalarField, VectorField,
                            TemporalVectorFields)):
        raise TypeError("Unknown type for 'obj' : {}".format(type(obj)))
    if not isinstance(wall_direction, NUMBERTYPES):
        raise TypeError("'wall_direction' must be a number")
    if wall_direction != 1 and wall_direction != 2:
        raise ValueError("'wall_direction' must be 1 or 2")
    if not isinstance(wall_position, NUMBERTYPES):
        raise ValueError("'wall_position' must be a number")
    axe_x, axe_y = obj.axe_x, obj.axe_y
    if interval is None:
        if wall_direction == 2:
            interval = [np.min(axe_x), np.max(axe_x)]
        else:
            interval = [np.min(axe_y), np.max(axe_y)]
    if not isinstance(interval, ARRAYTYPES):
        raise TypeError("'interval' must be a array")

    # Get data according to 'obj' type
    if isinstance(obj, ScalarField):
        # checking masked values
        if np.any(obj.mask):
            warnings.warn("I can give weird results if masked values remains")
        V = obj.values_as_sf
        if wall_direction == 1:
            axe = axe_x
        else:
            axe = axe_y
    elif isinstance(obj, VectorField):
        if np.any(obj.mask):
            warnings.warn("I can give weird results if masked values remains")
        if wall_direction == 1:
            V = obj.comp_y_as_sf
            axe = axe_x
        else:
            V = obj.comp_x_as_sf
            axe = axe_y
    elif isinstance(obj, TemporalVectorFields):
        if np.any(obj.fields[0].mask):
            warnings.warn("I can give weird results if masked values remains")
        pts = []
        times = obj.times
        if wall_direction == 1:
            unit_axe = obj.unit_y
        else:
            unit_axe = obj.unit_x
        for field in obj.fields:
            pts.append(get_separation_position(field,
                                               wall_direction=wall_direction,
                                               wall_position=wall_position,
                                               interval=interval))
        return Profile(times, pts, unit_x=obj.unit_times,
                       unit_y=unit_axe)
    else:
        raise ValueError("Unknown type for 'obj'")

    # Getting separation position
    # Getting lines around wall
    if wall_position < axe[0]:
        lines_pos = axe[0:nmb_lines]
    elif wall_position > axe[-1]:
        lines_pos = axe[-nmb_lines-1:-1]
    else:
        inds = V.get_indice_on_axe(wall_direction, wall_position)
        if len(inds) == 1:
            inds = [inds[0], inds[0] + 1]
        if inds[0] - nmb_lines/2 < 0:
            lines_pos = axe[0:nmb_lines]
        elif inds[-1] + nmb_lines/2 > len(axe):
            lines_pos = axe[-nmb_lines-1:-1]
        else:
            lines_pos = axe[inds[0] - nmb_lines/2:inds[1] + nmb_lines/2]
    # Getting separation points on surrounding lines
    seps = np.array([])
    new_lines_pos = np.array([])
    for lp in lines_pos:
        # extraction one line
        tmp_profile = V.get_profile(wall_direction, lp)
        # getting the velocity sign changment on the line
        values = tmp_profile.get_interpolated_values(y=0)
        values = np.array(values)
        # masking with 'interval'
        values = values[np.logical_and(values > interval[0],
                                       values < interval[1])]
        if len(values) == 0:
            continue
        seps = np.append(seps, np.mean(values))
        new_lines_pos = np.append(new_lines_pos, lp)
    if len(seps) == 0:
        raise Exception("Can't find sign chagment on the given interval")
    elif len(seps) != nmb_lines:
        warnings.warn("extrapolation done on only {} points"
                      " instead of {}".format(len(seps), nmb_lines))
    lines_pos = new_lines_pos
    # Deleting lines where no separation points were found
    if np.any(np.isnan(seps)):
        warnings.warn("I can't find a separation points on one (or more)"
                      " line(s). You may want to change 'interval' values")
        seps = seps[~np.isnan(seps)]
        lines_pos = lines_pos[~np.isnan(seps)]
    interp = UnivariateSpline(lines_pos, seps, k=1)
    return float(interp(wall_position))



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from .. import plotlib as pplt
import numpy as np
import unum
import copy
from ..utils import make_unit
from ..utils.types import ARRAYTYPES, INTEGERTYPES, NUMBERTYPES, STRINGTYPES
from . import points as pts


class TemporalPoints(object):
    """
    Class representing a set of time evolving points.
    """

    def __init__(self):
        self.point_sets = []
        self.__times = np.array([], dtype=float)
        self.__unit_times = make_unit("")
        self.unit_x = make_unit('')
        self.unit_y = make_unit('')
        self.unit_times = make_unit('')

    def __iter__(self):
        for i in np.arange(len(self.point_sets)):
            yield self.times[i], self.point_sets[i]

    def __len__(self):
        return len(self.point_sets)

    @property
    def unit_x(self):
        return self.__unit_x

    @unit_x.setter
    def unit_x(self, value):
        self.__unit_x = value
        for pt in self.point_sets:
            pt.unit_x = value

    @property
    def unit_y(self):
        return self.__unit_y

    @unit_y.setter
    def unit_y(self, value):
        self.__unit_y = value
        for pt in self.point_sets:
            pt.unit_y = value

    @property
    def times(self):
        return self.__times

    @times.setter
    def times(self, values):
        if not isinstance(values, ARRAYTYPES):
            raise TypeError()
        if len(self.point_sets) != len(values):
            raise ValueError("New number of time ({}) do not corespond to "
                             "the number of fields ({})"
                             .format(len(values), len(self.point_sets)))
        self.__times = values

    @times.deleter
    def times(self):
        raise Exception("Nope, can't do that")

    @property
    def dt(self):
        if len(self.times) > 1:
            return self.times[1] - self.times[0]
        else:
            return None

    @property
    def unit_times(self):
        return self.__unit_times

    @unit_times.setter
    def unit_times(self, new_unit_times):
        if isinstance(new_unit_times, unum.Unum):
            if new_unit_times.asNumber() == 1:
                self.__unit_times = new_unit_times
            else:
                raise ValueError()
        elif isinstance(new_unit_times, STRINGTYPES):
            self.__unit_times == make_unit(new_unit_times)
        else:
            raise TypeError()

    @unit_times.deleter
    def unit_times(self):
        raise Exception("Nope, can't do that")

    def scale(self, scalex=None, scaley=None, scalev=None, scalet=None,
              inplace=False):
        """
        Scale the point sets.

        Parameters
        ----------
        scalex, scaley, scalev : numbers or Unum objects
            Scale for the axis and the values.
        inplace : boolean
            .
        """
        if inplace:
            tmp_ps = self
        else:
            tmp_ps = self.copy()
        for i in range(len(self.point_sets)):
            tmp_ps.point_sets[i].scale(scalex=scalex,
                                       scaley=scaley,
                                       scalev=scalev,
                                       inplace=True)
        # scale the time
        if scalet is None:
            pass
        elif isinstance(scalet, NUMBERTYPES):
            tmp_ps.times *= scalet
        elif isinstance(scalet, unum.Unum):
            new_unit = tmp_ps.unit_times*scalet
            fact = new_unit.asNumber()
            new_unit /= fact
            tmp_ps.unit_times = new_unit
            tmp_ps.times *= fact
        else:
            raise TypeError()
        # returning
        return tmp_ps

    def change_unit(self, axe, new_unit):
        """
        Change the unit of an axe.

        Parameters
        ----------
        axe : string
            'y' for changing the profile y axis unit
            'x' for changing the profile x axis unit
            'values' for changing values unit
            'time' for changing time unit
        new_unit : Unum.unit object or string
            The new unit.
        """
        if isinstance(new_unit, STRINGTYPES):
            new_unit = make_unit(new_unit)
        if not isinstance(new_unit, unum.Unum):
            raise TypeError()
        if not isinstance(axe, STRINGTYPES):
            raise TypeError()
        if axe in ['x', 'y', 'values']:
            for pt in self.point_sets:
                pt.change_unit(axe, new_unit)
        elif axe == 'time':
            old_unit = self.unit_times
            new_unit = old_unit.asUnit(new_unit)
            fact = new_unit.asNumber()
            self.times *= fact
            self.unit_times = new_unit/fact
        else:
            raise ValueError()

    def add_pts(self, pt, time=0., unit_times=""):
        """
        Add a Pointsobject to the existing point set.

        Parameters
        ----------
        pt : Point object
            The points to add.
        time : number
            time associated to the field.
        unit_time : Unum object
            time unit.
        """
        if not isinstance(pt, pts.Points):
            raise TypeError()
        if not isinstance(time, NUMBERTYPES):
            raise TypeError("'time' should be a number, not {}"
                            .format(type(time)))
        if isinstance(unit_times, unum.Unum):
            if unit_times.asNumber() != 1:
                raise ValueError()
        elif isinstance(unit_times, STRINGTYPES):
            unit_times = make_unit(unit_times)
        else:
            raise TypeError()
        # if this is the first field
        if len(self.point_sets) == 0:
            self.unit_x = pt.unit_x
            self.unit_y = pt.unit_y
            self.unit_times = unit_times
            self.__times = np.asarray([time], dtype=float)
        # if not
        else:
            # storing time
            time = (time*self.unit_times/unit_times).asNumber()
            self.__times = np.append(self.__times, time)
        # use default constructor
        self.point_sets.append(pt)
        # sorting the field with time
        self.__sort_field_by_time()

    def remove_pts(self, indice):
        """
        Remove points of the existing point set.

        Parameters
        ----------
        indice : integer or list of integers
            Point indice(s) to remove.
        """
        if isinstance(indice, INTEGERTYPES):
            indice = [indice]
        for i in indice:
            self.__times = np.delete(self.times, i)
            self.point_sets = np.delete(self.point_sets, i)

    def crop(self, intervx=None, intervy=None, intervv=None, intervt=None,
             full_output=False, ind=False, inplace=False):
        """
        Return a croped field in respect with given intervals.

        Parameters
        ----------
        intervx : array, optional
            interval wanted along x
        intervy : array, optional
            interval wanted along y
        intervv : array, optional
            interval wanted along v
        intervt : array, optional
            interval wanted along time
        full_output : boolean, optional
            If 'True', cutting indices are also returned
        inplace : boolean, optional
            If 'True', fields are croped in place.
        """
        # check parameters
        if intervt is not None:
            if not isinstance(intervt, ARRAYTYPES):
                raise TypeError()
            intervt = np.array(intervt, dtype=float)
            if intervt.shape != (2, ):
                raise ValueError()
        # get wanted times
        if intervt is not None:
            if not ind:
                if intervt[0] < self.times[0]:
                    ind1 = 0
                elif intervt[0] > self.times[-1]:
                    raise ValueError()
                else:
                    ind1 = np.where(intervt[0] <= self.times)[0][0]
                if intervt[1] > self.times[-1]:
                    ind2 = len(self.times) - 1
                elif intervt[1] < self.times[0]:
                    raise ValueError()
                else:
                    ind2 = np.where(intervt[1] >= self.times)[0][-1]
                intervt = [ind1, ind2]
        # crop
        if inplace:
            croppts = self
        else:
            croppts = self.copy()
        # temporal
        if intervt is not None:
            croppts.point_sets = croppts.point_sets[intervt[0]:intervt[1] + 1]
            croppts.times = croppts.times[intervt[0]:intervt[1] + 1]
        # spatial
        for pt in self.point_sets:
            pt.crop(intervx=intervx, intervy=intervy, intervv=intervv,
                    inplace=True)
        # returning
        return croppts

    def smooth(self, tos='gaussian', size=None, inplace=False, **kw):
        """
        Return a smoothed points field set.

        Parameters
        ----------
        tos : string, optional
            Type of smoothing, can be 'uniform', 'gaussian' (default)
            or 'lowpass'.
        size : number, optional
            radius of the smoothing for 'uniform',
            radius of the smoothing for 'gaussian',
            cut off frequency for 'lowpass'
            Default are 3 for 'uniform',  1 for 'gaussian' and 0.1 for '
            lowpass'.
        inplace : boolean
            If 'False', return a smoothed points field
            else, smooth in place.
        kw : dic
            Additional parameters for ndimage methods
            (See ndimage documentation)
        """
        if inplace:
            tmp_pts = self
        else:
            tmp_pts = self.copy()
        for pt in tmp_pts.point_sets:
            pt.smooth(tos=tos, size=size, inplace=True, **kw)
        return tmp_pts

    def copy(self):
        """
        Return a copy of the velocityfields
        """
        return copy.deepcopy(self)

    def __sort_field_by_time(self):
        if len(self.point_sets) in [0, 1]:
            return None
        ind_sort = np.argsort(self.times)
        self.times = self.times[ind_sort]
        self.point_sets = [self.point_sets[i] for i in ind_sort]

    def display(self, sharecb=True, buffer_size=100, cpkw={}):
        """
        Create a windows to display temporals point sets, controlled by
        buttons.

        Parameters
        ----------
        sharecb: boolean
            Do all the vector field serie has to share the same colorbar or
            not.
        buffer_size: number
            Number of displays to keep in memory (faster, but use memory).
        cpkw: dict
            Drawing argument for points

        Display control
        ---------------
        The display can be controlled using the button, but also the keyboard:
        space, right arrow or + : next field
        backspace, left arrow or - : previous field
        up arrow : last field
        down arrow : first field
        number + enter : goto a specific frame
        p : play the animated fields
        number + i : set the animation increment
        number + t : set the animation time interval (ms)
        q : close
        s : save an image
        """
        # get data
        x = [list(pt.xy[:, 0]) for pt in self.point_sets]
        y = [list(pt.xy[:, 1]) for pt in self.point_sets]
        v = [list(pt.v) for pt in self.point_sets]
        # check if there is data
        if len(np.concatenate(x)) == 0:
            return None
        # default args
        args = {'kind': 'scatter', 'marker': 'o'}
        if 'kind' in list(cpkw.keys()):
            if cpkw['kind'] != 'scatter':
                args = {}
        args.update(cpkw)
        db = pplt.Displayer(x, y, values=v, buffer_size=buffer_size, **args)
        # plot
        if len(self.times) == 1:
            db.draw(0, rescale=False)
        else:
            pplt.ButtonManager(db)
        return db



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import matplotlib.pyplot as plt
from .. import plotlib as pplt
import numpy as np
import unum
import copy
from ..utils import make_unit, ProgressCounter
from ..utils.types import ARRAYTYPES, INTEGERTYPES, NUMBERTYPES, STRINGTYPES
from . import field as fld
from . import fields as flds
from . import scalarfield as sf
from . import points as pts
from . import profile as prof
from . import vectorfield as vf


class TemporalFields(flds.Fields, fld.Field):
    """
    Class representing a set of time evolving fields.
    All fields added to this object has to have the same axis system.
    """

    def __init__(self):
        fld.Field.__init__(self)
        flds.Fields.__init__(self)
        self.fields = []
        self.__times = np.array([], dtype=float)
        self.__unit_times = make_unit("")
        self.field_type = None
        self.axe_x = []
        self.axe_y = []
        self.unit_x = make_unit('')
        self.unit_y = make_unit('')
        self.unit_times = make_unit('')

    def __add__(self, other):
        if isinstance(other, self.fields[0].__class__):
            tmp_TF = self.copy()
            for i in np.arange(len(tmp_TF.fields)):
                tmp_TF.fields[i] += other
            return tmp_TF
        elif isinstance(other, self.__class__):
            tmp_tf = self.copy()
            if np.all(self.times == other.times):
                for i in np.arange(len(self.fields)):
                    tmp_tf.fields[i] += other.fields[i]
            else:
                for i in np.arange(len(other.fields)):
                    tmp_tf.add_field(other.fields[i])
            return tmp_tf
        else:
            tmp_TF = self.copy()
            for i in np.arange(len(tmp_TF.fields)):
                tmp_TF.fields[i] += other
            return tmp_TF
        # else:
        #     raise TypeError("cannot concatenate {} with"
        #                     " {}.".format(self.__class__, type(other)))

    def __sub__(self, other):
        return self.__add__(-other)

    def __neg__(self):
        tmp_tf = self.copy()
        for i in np.arange(len(self.fields)):
            tmp_tf.fields[i] = -tmp_tf.fields[i]
        return tmp_tf

    def __mul__(self, other):
        if isinstance(other, self.__class__):
            if not len(self) == len(other):
                raise Exception()
            if not np.all(self.axe_x == other.axe_x) \
                    and np.all(self.axe_y == other.axe_y):
                raise Exception()
            if not np.all(self.times == other.times):
                raise Exception()
            vfs = self.__class__()
            for i in np.arange(len(self.fields)):
                vfs.add_field(self.fields[i]*other.fields[i])
            return vfs
        elif isinstance(other, (NUMBERTYPES, unum.Unum)):
            final_vfs = self.__class__()
            for field in self.fields:
                final_vfs.add_field(field*other)
            return final_vfs
        else:
            raise TypeError("You can only multiply a temporal velocity field "
                            "by numbers")

    __rmul__ = __mul__

    def __truediv__(self, other):
        if isinstance(other, self.__class__):
            if not len(self) == len(other):
                raise Exception()
            if not np.all(self.axe_x == other.axe_x) \
                    and np.all(self.axe_y == other.axe_y):
                raise Exception()
            if not np.all(self.times == other.times):
                raise Exception()
            vfs = self.__class__()
            for i in np.arange(len(self.fields)):
                vfs.add_field(self.fields[i]/other.fields[i])
            return vfs
        elif isinstance(other, self.fields[0].__class__):
            if not np.all(self.axe_x == other.axe_x) \
                    and np.all(self.axe_y == other.axe_y):
                raise Exception()
            vfs = self.__class__()
            for i in np.arange(len(self.fields)):
                vfs.add_field(self.fields[i]/other)
            return vfs
        elif isinstance(other, (NUMBERTYPES, unum.Unum)):
            final_vfs = self.__class__()
            for i, field in enumerate(self.fields):
                final_vfs.add_field(other, time=self.times[i],
                                    unit_times=self.unit_times)
            return final_vfs
        else:
            raise TypeError("")

    __div__ = __truediv__

    def __pow__(self, number):
        if not isinstance(number, NUMBERTYPES):
            raise TypeError("You only can use a number for the power "
                            "on a Vectorfield")
        final_vfs = self.__class__()
        for field in self.fields:
            final_vfs.add_field(np.power(field, number))
        return final_vfs

    def __iter__(self):
        for i in np.arange(len(self.fields)):
            yield self.times[i], self.fields[i]

    def __eq__(self, obj):
        if not isinstance(obj, self.__class__):
            return False
        if not fld.Field.__eq__(self, obj):
            return False
        if not flds.Fields.__eq__(self, obj):
            return False
        for attr in ['fields', 'times', 'axe_x', 'axe_y']:
            if not np.all(self.__getattribute__(attr) ==
                          obj.__getattribute__(attr)):
                return False
        for attr in ['field_type', 'unit_x', 'unit_y', 'unit_times']:
            if not (self.__getattribute__(attr) ==
                    obj.__getattribute__(attr)):
                return False
        return True

    @fld.Field.axe_x.setter
    def axe_x(self, value):
        fld.Field.axe_x.fset(self, value)
        for field in self.fields:
            field.axe_x = value
            field.xy_scale = self.xy_scale

    @fld.Field.axe_y.setter
    def axe_y(self, value):
        fld.Field.axe_y.fset(self, value)
        for field in self.fields:
            field.axe_y = value
            field.xy_scale = self.xy_scale

    @fld.Field.unit_x.setter
    def unit_x(self, value):
        fld.Field.unit_x.fset(self, value)
        for field in self.fields:
            field.unit_x = value

    @fld.Field.unit_y.setter
    def unit_y(self, value):
        fld.Field.unit_y.fset(self, value)
        for field in self.fields:
            field.unit_y = value

    @property
    def mask(self):
        dim = (len(self.fields), self.shape[0], self.shape[1])
        mask_f = np.empty(dim, dtype=bool)
        for i, field in enumerate(self.fields):
            mask_f[i, :, :] = field.mask[:, :]
        return mask_f

    @property
    def mask_as_sf(self):
        dim = len(self.fields)
        mask_f = np.empty(dim, dtype=object)
        for i, field in enumerate(self.fields):
            mask_f[i] = field.mask_as_sf
        return mask_f

    @property
    def mask_cum(self):
        cum_mask = np.sum(self.mask, axis=0)
        cum_mask = cum_mask == len(self.mask)
        return cum_mask

    @property
    def mask_cum_as_sf(self):
        cum_mask = self.mask_cum
        tmp_sf = sf.ScalarField()
        tmp_sf.import_from_arrays(self.axe_x, self.axe_y, cum_mask, mask=None,
                                  unit_x=self.unit_x, unit_y=self.unit_y,
                                  unit_values='')
        return tmp_sf

    @property
    def times(self):
        return self.__times

    @times.setter
    def times(self, values):
        if not isinstance(values, ARRAYTYPES):
            raise TypeError()
        if len(self.fields) != len(values):
            raise ValueError("New number of time ({}) do not corespond to "
                             "the number of fields ({})"
                             .format(len(values), len(self.fields)))
        self.__times = values

    @times.deleter
    def times(self):
        raise Exception("Nope, can't do that")

    @property
    def dt(self):
        return self.times[1] - self.times[0]

    @property
    def unit_times(self):
        return self.__unit_times

    @unit_times.setter
    def unit_times(self, new_unit_times):
        if isinstance(new_unit_times, unum.Unum):
            if new_unit_times.asNumber() == 1:
                self.__unit_times = new_unit_times
            else:
                raise ValueError()
        elif isinstance(new_unit_times, STRINGTYPES):
            self.__unit_times == make_unit(new_unit_times)
        else:
            raise TypeError()

    @unit_times.deleter
    def unit_times(self):
        raise Exception("Nope, can't do that")

    @property
    def unit_values(self):
        if len(self.fields) != 0:
            return self[0].unit_values

    def get_mean_field(self, nmb_min=1, dtype=None):
        """
        Calculate the mean velocity field, from all the fields.

        Parameters
        ----------
        nmb_min : integer, optional
            Minimum number of values used to make a mean. else, the value is
            masked
        dtype : type
            Specify the output values type (default to the same one as fields).
        """
        # checks
        if len(self.fields) == 0:
            raise ValueError("There is no fields in this object")
        if self.field_type == vf.VectorField:
            value = [0., 0.]
        else:
            value = 0.
        if not dtype:
            dtype = self.fields[0]._values_dtype
        #
        result_f = self.fields[0].copy()
        result_f.change_dtype(float)
        result_f.fill(kind='value', value=value, crop=False, inplace=True)
        mask_cum = np.zeros(self.shape, dtype=int)
        mask_cum[np.logical_not(self.fields[0].mask)] += 1
        i = 0
        for field in self.fields[1::]:
            i += 1
#            print("{},  {}".format(i, field.unit_values))
            added_field = field.copy()
            added_field.fill(kind='value', value=0., inplace=True)
            result_f += added_field
            mask_cum[np.logical_not(field.mask)] += 1
        mask = mask_cum <= nmb_min
        result_f.mask = mask
        result_f.change_dtype(dtype)
        fact = mask_cum
        fact[mask] = 1
        result_f /= fact
        result_f.xy_scale = self.fields[0].xy_scale
        return result_f

    def get_interpolated_field(self, time):
        """
        Return the interpolated field happening at the time 'time'.
        """
        # check
        assert isinstance(time, NUMBERTYPES)
        assert time >= self.times[0]
        assert time <= self.times[-1]
        # if time is in self.times
        if np.any(self.times == time):
            return self.fields[self.times == time][0]
        # else, get the surrounding fields
        ind_time = np.argwhere(self.times > time)[0][0]
        denom = self.times[ind_time] - self.times[ind_time - 1]
        coef1 = (self.times[ind_time] - time)/denom
        coef2 = (time - self.times[ind_time - 1])/denom
        new_field = self.fields[ind_time]*coef2 + \
                    self.fields[ind_time - 1]*coef1
        new_field.time = time
        # returning
        assert isinstance(new_field, self.fields[0].__class__)
        return new_field

    def get_fluctuant_fields(self, nmb_min_mean=1):
        """
        Calculate the fluctuant fields (fields minus mean field).

        Parameters
        ----------
        nmb_min_mean : number, optional
            Parameter for mean computation (see 'get_mean_field' doc).

        Returns
        -------
        fluct_fields : TemporalScalarFields or TemporalVectorFields object
            Contening fluctuant fields.
        """
        fluct_fields = self.__class__()
        mean_field = self.get_mean_field(nmb_min=nmb_min_mean)
        for i, field in enumerate(self.fields):
            fluct_fields.add_field(field - mean_field, time=self.times[i],
                                   unit_times=self.unit_times)
        return fluct_fields

    def get_spatial_spectrum(self, component, direction, intervx=None,
                             intervy=None, intervtime=None, welch_seglen=None,
                             scaling='base', fill='linear'):
        """
        Return a spatial spectrum.
        If more than one time are specified, spectrums are averaged.

        Parameters
        ----------
        component : string
            Should be an attribute name of the stored fields.
        direction : string
            Direction in which perform the spectrum ('x' or 'y').
        intervx and intervy : 2x1 arrays of number, optional
            To chose the zone where to calculate the spectrum.
            If not specified, the biggest possible interval is choosen.
        intervtime : 2x1 array, optional
            Interval of time on which averaged the spectrum.
        welch_seglen : integer, optional
            If specified, welch's method is used (dividing signal into
            overlapping segments, and averaging periodogram) with the given
            segments length (in number of points).
        scaling : string, optional
            If 'base' (default), result are in component unit.
            If 'spectrum', the power spectrum is returned (in unit^2).
            If 'density', the power spectral density is returned
            (in unit^2/(1/unit_axe))
        fill : string or float
            Specifies the way to treat missing values.
            A value for value filling.
            A string ('linear', 'nearest' or 'cubic') for interpolation.

        Notes
        -----
        If there is missing values on the field, 'fill' is used to linearly
        interpolate the missing values (can impact the spectrum).
        """
        # check parameters
        try:
            self[0].__getattribute__('{}_as_sf'.format(component))
        except AttributeError():
            raise ValueError()
        if not isinstance(direction, STRINGTYPES):
            raise TypeError()
        if direction not in ['x', 'y']:
            raise ValueError()
        if intervtime is None:
            intervtime = [self.times[0], self.times[-1]]
        if not isinstance(intervtime, ARRAYTYPES):
            raise TypeError()
        intervtime = np.array(intervtime)
        if not intervtime.shape == (2,):
            raise ValueError()
        if intervtime[0] < self.times[0]:
            intervtime[0] = self.times[0]
        if intervtime[-1] > self.times[-1]:
            intervtime[-1] = self.times[-1]
        if intervtime[0] >= intervtime[1]:
            raise ValueError()
        # loop on times
        spec = 0
        nmb = 0
        for i, time in enumerate(self.times):
            if time < intervtime[0] or time > intervtime[1]:
                continue
            comp = self[i].__getattribute__('{}_as_sf'.format(component))
            if spec == 0:
                spec = comp.get_spatial_spectrum(direction, intervx=intervx,
                                                 intervy=intervy,
                                                 welch_seglen=welch_seglen,
                                                 scaling=scaling, fill=fill)
            else:
                spec += comp.get_spatial_spectrum(direction, intervx=intervx,
                                                  intervy=intervy,
                                                  welch_seglen=welch_seglen,
                                                  scaling=scaling, fill=fill)
            nmb += 1
        # returning
        spec /= nmb
        return spec

    def get_time_profile(self, component, pt, wanted_times=None, ind=False):
        """
        Return a profile contening the time evolution of the given component.

        Parameters
        ----------
        component : string
            Should be an attribute name of the stored fields.
        pt : 2x1 array of numbers, or pts.Points object
            Wanted position for the time profile, in axis units.
        wanted_times : 2x1 array of numbers
            Time interval in which getting profile (default is all).
        ind : boolean, optional
            If 'True', values are undersood as indices.

        Returns
        -------
        profile : prof.Profile object

        """
        # check parameters coherence
        if not isinstance(component, STRINGTYPES):
            raise TypeError("'component' must be a string")
        if isinstance(pt, ARRAYTYPES):
            if ind:
                if pt[0] % 1 != 0 or pt[1] % 1 != 0:
                    raise ValueError()
                ind_x = int(pt[0])
                ind_y = int(pt[1])
            else:
                ind_x = self.get_indice_on_axe(1, pt[0], kind='nearest')
                ind_y = self.get_indice_on_axe(2, pt[1], kind='nearest')
            axe_x, axe_y = self.axe_x, self.axe_y
            if not (0 <= ind_x < len(axe_x) and 0 <= ind_y < len(axe_y)):
                raise ValueError("'x' ans 'y' values out of bounds")
            pt = np.array([[ind_x, ind_y]]*len(self.times), dtype=int)
            mask_times = np.zeros(len(self.times), dtype=bool)
        if isinstance(pt, pts.Points):
            mask_times = [time not in pt.v for time in self.times]
            mask_times = np.array(mask_times, dtype=bool)
            pt = [[self.get_indice_on_axe(1, pt.xy[i, 0], kind='nearest'),
                   self.get_indice_on_axe(2, pt.xy[i, 1], kind='nearest')]
                  for i in range(len(pt.xy))]
            pt = np.array(pt, dtype=int)
        if wanted_times is not None:
            if wanted_times[-1] <= wanted_times[0]:
                raise ValueError()
            mask_times = np.logical_or(self.times < wanted_times[0],
                                       mask_times)
            mask_times = np.logical_or(self.times > wanted_times[1],
                                       mask_times)
        # getting wanted time if necessary
        w_times_ind = np.arange(len(self.times))[~mask_times]
        # getting component values
        dim = len(w_times_ind)
        compo = np.empty(dim, dtype=float)
        masks = np.empty(dim, dtype=float)
        for i, time_ind in enumerate(w_times_ind):
            ind_x, ind_y = pt[i]
            compo[i] = self.fields[time_ind].__getattribute__(component)[ind_x,
                                                                         ind_y]
            masks[i] = self.fields[time_ind].mask[ind_x, ind_y]
        # gettign others datas
        time = self.times[w_times_ind]
        unit_time = self.unit_times
        unit_values = self.unit_values
        # getting position indices
        return prof.Profile(time, compo, masks, unit_x=unit_time,
                            unit_y=unit_values)

    def inject_time_profile(self, comp, pt, prof, ind=False):
        """
        Overwrite the value at the given points with data from the time
        profile.

        Parameters
        ----------
        comp : string
            Should be an attribute name of the stored fields.
        pt : 2x1 array of numbers, or pts.Points object
            Wanted position for the time profile, in axis units.
        prof : Profile object
            Profile used to overwrite data at point
        ind : boolean, optional
            If 'True', values are understood as indices.
        """
        # check
        if not isinstance(comp, STRINGTYPES):
            raise TypeError("'comp' must be a string")
        if isinstance(pt, ARRAYTYPES):
            if ind:
                if pt[0] % 1 != 0 or pt[1] % 1 != 0:
                    raise ValueError()
                ind_x = int(pt[0])
                ind_y = int(pt[1])
            else:
                ind_x = self.get_indice_on_axe(1, pt[0], kind='nearest')
                ind_y = self.get_indice_on_axe(2, pt[1], kind='nearest')
            axe_x, axe_y = self.axe_x, self.axe_y
            if not (0 <= ind_x < len(axe_x) and 0 <= ind_y < len(axe_y)):
                raise ValueError("'x' ans 'y' values out of bounds")
            pt = np.array([[ind_x, ind_y]]*len(self.times), dtype=int)
            mask_times = np.zeros(len(self.times), dtype=bool)
        if isinstance(pt, pts.Points):
            mask_times = [time not in pt.v for time in self.times]
            mask_times = np.array(mask_times, dtype=bool)
            pt = [[self.get_indice_on_axe(1, pt.xy[i, 0], kind='nearest'),
                   self.get_indice_on_axe(2, pt.xy[i, 1], kind='nearest')]
                  for i in range(len(pt.xy))]
            pt = np.array(pt, dtype=int)
        if len(prof) != len(self):
            raise ValueError()
        # do it
        for i in range(len(prof)):
            self.fields[i].__getattribute__(comp)[ind_x, ind_y] = prof.y[i]

    def get_temporal_spectrum(self, component, pt, ind=False,
                              wanted_times=None, welch_seglen=None,
                              scaling='base', fill='linear', mask_error=True,
                              detrend='constant'):
        """
        Return a Profile object, with the temporal spectrum of
        'component' on the point 'pt'.

        Parameters
        ----------
        component : string
            .
        pt : 2x1 array of numbers
            .
        ind : boolean
            If true, 'pt' is read as indices,
            else, 'pt' is read as coordinates.
        wanted_times : 2x1 array, optional
            Time interval in which compute spectrum (default is all).
        welch_seglen : integer, optional
            If specified, welch's method is used (dividing signal into
            overlapping segments, and averaging periodogram) with the given
            segments length (in number of points).
        scaling : string, optional
            If 'base' (default), result are in component unit.
            If 'spectrum', the power spectrum is returned (in unit^2).
            If 'density', the power spectral density is returned (in unit^2/Hz)
        fill : string or float
            Specifies the way to treat missing values.
            A value for value filling.
            A string ('linear', 'nearest', 'zero', 'slinear', 'quadratic,
            'cubic' where 'slinear', 'quadratic' and 'cubic' refer to a spline
            interpolation of first, second or third order) for interpolation.
        mask_error : boolean
            If 'False', instead of raising an error when masked value appear on
            time profile, '(None, None)' is returned.
        detrend : string, optional
            Method used to detrend the profile. Can be 'none',
            'constant' (default) or 'linear'.

        Returns
        -------
        magn_prof : prof.Profile object
            Magnitude spectrum.
        """
        # checking parameters coherence
        if not isinstance(pt, ARRAYTYPES):
            raise TypeError("'pt' must be a 2x1 array")
        if ind:
            pt = np.array(pt, dtype=int)
        else:
            pt = np.array(pt, dtype=float)
        if not pt.shape == (2,):
            raise ValueError("'pt' must be a 2x1 array")
        if ind and (not isinstance(pt[0], INTEGERTYPES) or
                    not isinstance(pt[1], INTEGERTYPES)):
            raise TypeError("If 'ind' is True, 'pt' must be an array of two"
                            " integers")
        if not isinstance(ind, bool):
            raise TypeError("'ind' must be a boolean")
        x = pt[0]
        y = pt[1]
        # getting time profile
        time_prof = self.get_time_profile(component, pt=[x, y], ind=ind,
                                          wanted_times=wanted_times)
        magn_prof = time_prof.get_spectrum(welch_seglen=welch_seglen,
                                           scaling=scaling, fill=fill,
                                           mask_error=mask_error,
                                           detrend=detrend)
        return magn_prof

    def get_temporal_spectrum_over_area(self, component, intervx, intervy,
                                        ind=False, welch_seglen=None,
                                        scaling='base', fill='linear',
                                        detrend='constant'):
        """
        Return a prof.Profile object, contening a mean spectrum of the given
        component, on all the points included in the given intervals.

        Parameters
        ----------
        component : string
            Scalar component ('Vx', 'Vy', 'magnitude', ...).
        intervx, intervy : 2x1 arrays of numbers
            Defining the square on which averaging the spectrum.
            (in axes values)
        ind : boolean
            If true, 'pt' is read as indices,
            else, 'pt' is read as coordinates.
        welch_seglen : integer, optional
            If specified, welch's method is used (dividing signal into
            overlapping segments, and averaging periodogram) with the given
            segments length (in number of points).
        scaling : string, optional
            If 'base' (default), result are in component unit.
            If 'spectrum', the power spectrum is returned (in unit^2).
            If 'density', the power spectral density is returned (in unit^2/Hz)
        fill : string or float
            Specifies the way to treat missing values.
            A value for value filling.
            A string ('linear', 'nearest', 'zero', 'slinear', 'quadratic,
            'cubic' where 'slinear', 'quadratic' and 'cubic' refer to a spline
            interpolation of first, second or third order) for interpolation.
        detrend : string, optional
            Method used to detrend the profile. Can be 'none',
            'constant' (default) or 'linear'.

        Returns
        -------
        magn_prof : prof.Profile object
            Averaged magnitude spectrum.
        """
        # checking parameters coherence
        if not isinstance(component, STRINGTYPES):
            raise TypeError("'component' must be a string")
        if not isinstance(intervx, ARRAYTYPES):
            raise TypeError("'intervx' must be an array")
        if not isinstance(intervy, ARRAYTYPES):
            raise TypeError("'intervy' must be an array")
        if not isinstance(intervx[0], NUMBERTYPES):
            raise TypeError("'intervx' must be an array of numbers")
        if not isinstance(intervy[0], NUMBERTYPES):
            raise TypeError("'intervy' must be an array of numbers")
        axe_x, axe_y = self.axe_x, self.axe_y
        # checking interval values and getting bound indices
        if ind:
            if not isinstance(intervx[0], int)\
                    or not isinstance(intervx[1], int)\
                    or not isinstance(intervy[0], int)\
                    or not isinstance(intervy[1], int):
                raise TypeError("'intervx' and 'intervy' must be arrays of"
                                " integer if 'ind' is 'True'")
            if intervx[0] < 0 or intervy[0] < 0\
                    or intervx[-1] >= len(axe_x)\
                    or intervy[-1] >= len(axe_y):
                raise ValueError("intervals are out of bounds")
            ind_x_min = intervx[0]
            ind_x_max = intervx[1]
            ind_y_min = intervy[0]
            ind_y_max = intervy[1]
        else:
            axe_x_min = np.min(axe_x)
            axe_x_max = np.max(axe_x)
            axe_y_min = np.min(axe_y)
            axe_y_max = np.max(axe_y)
            if np.min(intervx) < axe_x_min\
                    or np.max(intervx) > axe_x_max\
                    or np.min(intervy) < axe_y_min\
                    or np.max(intervy) > axe_y_max:
                raise ValueError("intervals ({}) are out of bounds ({})"
                                 .format([intervx, intervy],
                                         [[axe_x_min, axe_x_max],
                                          [axe_y_min, axe_y_max]]))
            ind_x_min = self.get_indice_on_axe(1, intervx[0])[-1]
            ind_x_max = self.get_indice_on_axe(1, intervx[1])[0]
            ind_y_min = self.get_indice_on_axe(2, intervy[0])[-1]
            ind_y_max = self.get_indice_on_axe(2, intervy[1])[0]
        # Averaging ponctual spectrums
        magn = 0.
        nmb_fields = (ind_x_max - ind_x_min + 1)*(ind_y_max - ind_y_min + 1)
        real_nmb_fields = nmb_fields
        for i in np.arange(ind_x_min, ind_x_max + 1):
            for j in np.arange(ind_y_min, ind_y_max + 1):
                tmp_m = self.get_temporal_spectrum(component, [i, j], ind=True,
                                                   welch_seglen=welch_seglen,
                                                   scaling=scaling,
                                                   fill=fill, mask_error=True,
                                                   detrend=detrend)
                # check if the position is masked
                if tmp_m is None:
                    real_nmb_fields -= 1
                else:
                    magn = magn + tmp_m
        if real_nmb_fields == 0:
            raise Exception("I can't find a single non-masked time profile"
                            ", maybe you will want to try 'zero_fill' "
                            "option")
        magn = magn/real_nmb_fields
        return magn

    def get_spectrum_map(self, comp, welch_seglen=None, nmb_pic=1,
                         spec_smooth=None,
                         verbose=True):
        """
        Return the temporal spectrum map.

        Parameters
        ----------
        comp : string
            Component to get the spectrum from.
        welch_seglen : integer
            .
        nmb_pic : integer
            Number of succesive spectrum pic to detect
        spec_smooth : number
            .
        verbose : bool
            .

        Returns
        -------
        map_freq_sf :
            .
        map_freq_quality_sf :
            .
        """
        # check
        try:
            self.fields[0].__getattribute__(comp)
        except AttributeError():
            raise ValueError()
        # prepare
        map_freq = []
        map_freq_quality = []
        map_freq_mask = np.zeros(self.shape, dtype=bool)
        for i in range(nmb_pic):
            map_freq.append(np.zeros(self.shape, dtype=float))
            map_freq_quality.append(np.zeros(self.shape, dtype=float))
        if verbose:
            PG = ProgressCounter(init_mess="Begin spectrum map computation on {}"
                                 .format(comp),
                                 nmb_max=self.shape[0]*self.shape[1],
                                 name_things="points")
        # loop on field points
        for i, x in enumerate(self.axe_x):
            for j, y in enumerate(self.axe_y):
                if verbose:
                    PG.print_progress()
                # get local spectrum
                tmp_prof = self.get_time_profile(comp, [x, y])
                # check if should be masked
                if np.sum(tmp_prof.mask)/float(len(tmp_prof)) > .5:
                    map_freq_mask[i, j] = True
                    continue
                spec = tmp_prof.get_spectrum(welch_seglen=welch_seglen)
                # smooth if necessary
                if spec_smooth is not None:
                    spec.smooth(tos='gaussian', size=spec_smooth, inplace=True)
                # get maximale frequences
                for n in range(nmb_pic):
                    spec_max = spec.max
                    max_pos_ind = spec.get_value_position(spec_max,
                                                          ind=True)[0]
                    map_freq[n][i, j] = spec.x[max_pos_ind]
                    # get spectrum 'quality'
                    filt = np.logical_not(spec.mask)
                    spec_var = np.mean((spec.y[filt] -
                                       np.mean(spec.y[filt]))**2)**.5
                    map_freq_quality[n][i, j] = (spec_max - spec.mean)/spec_var
                    # remove this particular pic
                    spec.mask[max_pos_ind] = True
        # store results
        maps_freq = []
        maps_freq_quality = []
        for i in range(nmb_pic):
            map_freq_sf = sf.ScalarField()
            map_freq_sf.import_from_arrays(axe_x=self.axe_x, axe_y=self.axe_y,
                                           values=map_freq[i],
                                           mask=map_freq_mask,
                                           unit_x=self.unit_x,
                                           unit_y=self.unit_y,
                                           unit_values=spec.unit_y)
            map_freq_quality_sf = sf.ScalarField()
            map_freq_quality_sf.import_from_arrays(axe_x=self.axe_x,
                                                   axe_y=self.axe_y,
                                                   values=map_freq_quality[i],
                                                   mask=map_freq_mask,
                                                   unit_x=self.unit_x,
                                                   unit_y=self.unit_y,
                                                   unit_values='')
            maps_freq.append(map_freq_sf)
            maps_freq_quality.append(map_freq_quality_sf)
        # return
        if nmb_pic == 1:
            return maps_freq[0], maps_freq_quality[0]
        else:
            return maps_freq, maps_freq_quality

    def _get_comp_spectral_filtering(self, comp, fmin, fmax, order=2):
        """
        Perform a temporal spectral filtering on the wanted component

        Parameters
        ----------
        fmin, fmax : numbers
            Minimal and maximal frequencies
        order : integer, optional
            Butterworth filter order

        Returns
        -------
        filt_tf : TemporalFields
            Filtered temporal field
        """
        # check
        try:
            tf = self.__getattribute__("{}_as_sf".format(comp))
        except AttributeError():
            raise ValueError()
        # loop on space
        xys = [[x, y]
               for x in tf.axe_x
               for y in tf.axe_y]
        for x, y in xys:
            tmp_prof = tf.get_time_profile('values', [x, y])
            filt_tmp_prof = tmp_prof.spectral_filtering(fmin, fmax,
                                                        order=order)
            tf.inject_time_profile('values', [x, y], filt_tmp_prof)
        return tf

    def get_recurrence_map(self, norm=2, verbose=False, bandwidth=None,
                           normalized=False):
        """
        Return the recurrence map associated with the 2-norm.

        Returns
        -------
        rec_map : sf.ScalarField object
            .
        """
        length = len(self.fields)
        rec_map = np.zeros((length, length))
        if verbose:
            pc = ProgressCounter(init_mess="Begin recurence map computation",
                                 nmb_max=int(length**2/2. + length/2.),
                                 name_things='norms', perc_interv=1)
        if self.field_type == vf.VectorField:
            field_type = 0
            nmb_val = self.shape[0]*self.shape[1]*2
        elif self.field_type == sf.ScalarField:
            field_type = 1
            nmb_val = self.shape[0]*self.shape[1]
        else:
            raise Exception()
        inv_norm = 1./norm
        if norm % 2 == 0:
            need_abs = False
        else:
            need_abs = True
        # norm
        if normalized:
            if field_type == 0:
                magnitude = self.get_mean_field().magnitude
                normalizer = np.sum(magnitude**norm)**inv_norm/nmb_val
            else:
                magnitude = self.get_mean_field().values
                normalizer = np.sum(magnitude**norm)**inv_norm/nmb_val
        # double loop
        for i in range(length):
            for j in range(length):
                if i > j:
                    continue
                if i == j:
                    rec_map[i, j] = 0
                    continue
                if bandwidth is not None:
                    if np.abs(i - j) > bandwidth:
                        rec_map[i, j] = np.nan
                        rec_map[j, i] = np.nan
                        continue
                if verbose:
                    pc.print_progress()
                # get difference fields
                if field_type == 0:
                    valuesx = self.fields[i].comp_x - self.fields[j].comp_x
                    valuesy = self.fields[i].comp_y - self.fields[j].comp_y
                    values = np.concatenate((valuesx, valuesy))
                elif field_type == 1:
                    values = self.fields[i].values - self.fields[j].values
                # get norm
                if need_abs:
                    res = np.sum(np.abs(values)**norm)**inv_norm/nmb_val
                else:
                    res = np.sum(values**norm)**inv_norm/nmb_val
                rec_map[i, j] = res
                rec_map[j, i] = res
        # normalize
        if normalized:
            rec_map /= normalizer
        # return
        rec_map_sf = sf.ScalarField()
        rec_map_sf.import_from_arrays(axe_x=self.times, axe_y=self.times,
                                      values=rec_map,
                                      unit_x=self.unit_times,
                                      unit_y=self.unit_times,
                                      unit_values=self.unit_values)
        return rec_map_sf

    def extend(self, nmb_left=0, nmb_right=0, nmb_up=0, nmb_down=0,
               inplace=False):
        """
        Add columns or lines of masked values at the fields.

        Parameters
        ----------
        nmb_**** : integers
            Number of lines/columns to add in each direction.
        inplace : bool
            If 'False', return a new extended field, if 'True', modify the
            field inplace.
        Returns
        -------
        Extended_field : TemporalFields object, optional
            Extended field.
        """
        if inplace:
            tmp_tf = self
        else:
            tmp_tf = self.copy()
        # scale axis
        fld.Field.extend(tmp_tf, nmb_left=nmb_left, nmb_right=nmb_right,
                         nmb_up=nmb_up, nmb_down=nmb_down, inplace=True)
        # scale fields
        for i, _ in enumerate(tmp_tf.fields):
            tmp_tf.fields[i].extend(nmb_left=nmb_left, nmb_right=nmb_right,
                                    nmb_up=nmb_up, nmb_down=nmb_down,
                                    inplace=True)
        # return
        if not inplace:
            return tmp_tf

    def scale(self, scalex=None, scaley=None, scalev=None, scalet=None,
              inplace=False):
        """
        Scale the Fields.

        Parameters
        ----------
        scalex, scaley, scalev : numbers or Unum objects
            Scale for the axis and the values.
        inplace : boolean
            .
        """
        if inplace:
            tmp_f = self
        else:
            tmp_f = self.copy()
        # scale the field (automaticly scale the fields axis)
        # for compatibility purpose
        if not hasattr(self, "xy_scale"):
            self.xy_scale = make_unit('')
        for field in self.fields:
            if not hasattr(field, "xy_scale"):
                field.xy_scale = self.xy_scale
        fld.Field.scale(tmp_f, scalex=scalex, scaley=scaley,
                        inplace=True)
        # scale the values
        flds.Fields.scale(tmp_f, scalev=scalev, inplace=True)
        # scale the time
        if scalet is None:
            pass
        elif isinstance(scalet, NUMBERTYPES):
            tmp_f.times *= scalet
        elif isinstance(scalet, unum.Unum):
            new_unit = tmp_f.unit_times*scalet
            fact = new_unit.asNumber()
            new_unit /= fact
            tmp_f.unit_times = new_unit
            tmp_f.times *= fact
        else:
            raise TypeError()
        # returning
        if not inplace:
            return tmp_f

    def make_evenly_spaced(self, interp='linear', res=1):
        """
        Use interpolation to make the fields evenly spaced

        Parameters
        ----------
        interp : {‘linear’, ‘cubic’, ‘quintic’}, optional
            The kind of spline interpolation to use. Default is ‘linear’.
        res : number
            Resolution of the resulting field.
            A value of 1 meaning a spatial resolution equal to the smallest
            space along the two axis for the initial field.
        """
        # raise NotImplementedError('Not implemented yet')
        for field in self.fields:
            field.make_evenly_spaced(interp=interp, res=res)
        self._Field__axe_x = self.fields[0].axe_x
        self._Field__axe_y = self.fields[0].axe_y
        self.__mask = False

    def change_unit(self, axe, new_unit):
        """
        Change the unit of an axe.

        Parameters
        ----------
        axe : string
            'y' for changing the profile y axis unit
            'x' for changing the profile x axis unit
            'values' for changing values unit
            'time' for changing time unit
        new_unit : Unum.unit object or string
            The new unit.
        """
        if isinstance(new_unit, STRINGTYPES):
            new_unit = make_unit(new_unit)
        if not isinstance(new_unit, unum.Unum):
            raise TypeError()
        if not isinstance(axe, STRINGTYPES):
            raise TypeError()
        if axe in ['x', 'y', 'values']:
            for field in self.fields:
                field.change_unit(axe, new_unit)
        elif axe == 'time':
            old_unit = self.unit_times
            new_unit = old_unit.asUnit(new_unit)
            fact = new_unit.asNumber()
            self.times *= fact
            self.unit_times = new_unit/fact
        else:
            raise ValueError()
        if axe in ['x', 'y']:
            fld.Field.change_unit(self, axe, new_unit)

    def add_field(self, field, time=0., unit_times="", copy=True):
        """
        Add a field to the existing fields.

        Parameters
        ----------
        field : vf.VectorField or sf.ScalarField object
            The field to add.
        time : number
            time associated to the field.
        unit_time : Unum object
            time unit.
        """
        # TODO : pas de vérification de la cohérence des unités !
        # checking parameters
        if not isinstance(field, (vf.VectorField, sf.ScalarField)):
            raise TypeError()
        if self.field_type is None:
            self.field_type = field.__class__
        if not isinstance(field, self.field_type):
            raise TypeError()
        if not isinstance(time, NUMBERTYPES):
            raise TypeError("'time' should be a number, not {}"
                            .format(type(time)))
        if isinstance(unit_times, unum.Unum):
            if unit_times.asNumber() != 1:
                raise ValueError()
        elif isinstance(unit_times, STRINGTYPES):
            unit_times = make_unit(unit_times)
        else:
            raise TypeError()
        # if this is the first field
        if len(self.fields) == 0:
            self.axe_x = field.axe_x
            self.axe_y = field.axe_y
            self.unit_x = field.unit_x
            self.unit_y = field.unit_y
            self.unit_times = unit_times
            self.__times = np.asarray([time], dtype=float)
            self.field_type = field.__class__
        # if not
        else:
            # checking field type
            if not isinstance(field, self.field_type):
                raise TypeError()
            # checking axis
            axe_x, axe_y = self.axe_x, self.axe_y
            vaxe_x, vaxe_y = field.axe_x, field.axe_y
            if not np.all(axe_x == vaxe_x) and np.all(axe_y == vaxe_y):
                raise ValueError("Axes of the new field must be consistent "
                                 "with current axes")
            # storing time
            time = (time*self.unit_times/unit_times).asNumber()
            self.__times = np.append(self.__times, time)
        # use default constructor
        flds.Fields.add_field(self, field, copy=copy)
        # sorting the field with time
        self.__sort_field_by_time()

    def remove_fields(self, fieldnumbers):
        """
        Remove field(s) of the existing fields.

        Parameters
        ----------
        fieldnumber : integer or list of integers
            Velocity field(s) number(s) to remove.
        """
        if isinstance(fieldnumbers, INTEGERTYPES):
            fieldnumbers = [fieldnumbers]
        for nmb in fieldnumbers:
            self.__times = np.delete(self.times, nmb)
        flds.Fields.remove_field(self, fieldnumbers)

    def reduce_spatial_resolution(self, fact, inplace=False, verbose=False):
        """
        Reduce the spatial resolution of the fields by a factor 'fact'

        Parameters
        ----------
        fact : int
            Reducing factor.
        inplace : boolean, optional
            .
        """
        if inplace:
            tmpfs = self
        else:
            tmpfs = self.copy()
        # verbose
        if verbose:
            pg = ProgressCounter(f"Reducing resolution by {fact}",
                                 len(self), name_things='fields')
        #
        for f in tmpfs.fields:
            f.reduce_spatial_resolution(fact=fact, inplace=True)
            if verbose:
                pg.print_progress()
        #
        self.axe_x = self.fields[0].axe_x
        self.axe_y = self.fields[0].axe_y
        return tmpfs

    def reduce_temporal_resolution(self, nmb_in_interval, mean=True,
                                   inplace=False):
        """
        Return a TemporalVelocityFields, contening one field for each
        'nmb_in_interval' field in the initial TFVS.

        Parameters
        ----------
        nmb_in_interval : integer
            Length of the interval.
            (one field is kept for each 'nmb_in_interval fields)
        mean : boolean, optional
            If 'True', the resulting fields are average over the interval.
            Else, fields are taken directly.
        inplace : boolean, optional

        Returns
        -------
        TVFS : TemporalVelocityFields
        """
        # cehck parameters
        if not isinstance(nmb_in_interval, int):
            raise TypeError("'nmb_in_interval' must be an integer")
        if nmb_in_interval <= 0:
            raise TypeError("'nmb_in_interval' must be a positive integer")

        if nmb_in_interval == 1:
            return self.copy()
        if nmb_in_interval >= len(self):
            raise ValueError("'nmb_in_interval' is too big")
        if not isinstance(mean, bool):
            raise TypeError("'mean' must be a boolean")
        #
        tmp_TFS = self.__class__()
        i = 0
        times = self.times
        while True:
            tmp_f = self[i]
            time = times[i]
            if mean:
                for j in np.arange(i + 1, i + nmb_in_interval):
                    tmp_f += self[j]
                    time += times[j]
                tmp_f /= nmb_in_interval
                time /= nmb_in_interval
            tmp_TFS.add_field(tmp_f, time, self.unit_times)
            i += nmb_in_interval
            if i + nmb_in_interval >= len(self):
                break
        # returning
        if inplace:
            self.fields = tmp_TFS.fields
            self.times = tmp_TFS.times
        else:
            return tmp_TFS

    def augment_temporal_resolution(self, fact=2, inplace=False):
        """
        Augment the temporal resolution using temporal interpoalation.

        Parameters
        ----------
        fact : integer
            Temporal resolution ratio.
        inplace : bool
            .
        """
        # check
        assert type(fact) in [int], "TypeError"
        assert fact > 0, "ValueError"
        assert type(inplace) == bool, "TypeError"
        # fact = 1 (fool...)
        if fact == 1:
            if inplace:
                return None
            else:
                return self.copy()
        # get data
        if inplace:
            tf = self
        else:
            tf = self.copy()
        # get new times
        new_times = []
        for i in range(len(tf.times) - 1):
            tmp_times = np.linspace(tf.times[i], tf.times[i + 1],
                                    fact + 1)[0:-1]
            new_times.append(tmp_times)
        new_times = np.array(new_times).flatten()
        new_times = np.append(new_times, tf.times[-1])
        # loop on new times
        new_fields = []
        for time in new_times:
            new_fields.append(tf.get_interpolated_field(time))
        # store
        tf.fields = new_fields
        tf.times = new_times
        # returning
        if not inplace:
            return tf

    def crop_masked_border(self, hard=False, inplace=False):
        """
        Crop the masked border of the velocity fields in place.

        Parameters
        ----------
        hard : boolean, optional
            If 'True', partially masked border are croped as well.
        inplace : boolean, optional
            If 'True', crop the F in place,
            else, return a croped TF.
        """
        #get cumulated mask
        mask_cum = self.mask_cum
        # checking masked values presence
        if not np.any(mask_cum):
            return None
        # hard cropping
        if hard:
            if inplace:
                tmp_tf = self
            else:
                tmp_tf = self.copy()
            # remove trivial borders
            tmp_tf.crop_masked_border(hard=False, inplace=True)
            # until there is no more masked values
            while True:
                # getting mask
                masks = tmp_tf.mask
                mask = np.sum(masks, axis=0)
                mask = mask == len(tmp_tf.fields)
                # getting number of masked value on each border
                bd1 = np.sum(mask[0, :])
                bd2 = np.sum(mask[-1, :])
                bd3 = np.sum(mask[:, 0])
                bd4 = np.sum(mask[:, -1])
                # getting more masked border
                more_masked = np.argmax([bd1, bd2, bd3, bd4])
                # check remaining masked values
                if [bd1, bd2, bd3, bd4][more_masked] == 0:
                    break
                # deleting more masked border
                if more_masked == 0:
                    len_x = len(tmp_tf.axe_x)
                    tmp_tf.crop(intervx=[1, len_x], ind=True, inplace=True)
                elif more_masked == 1:
                    len_x = len(tmp_tf.axe_x)
                    tmp_tf.crop(intervx=[0, len_x - 2], ind=True,
                                inplace=True)
                elif more_masked == 2:
                    len_y = len(tmp_tf.axe_y)
                    tmp_tf.crop(intervy=[1, len_y], ind=True, inplace=True)
                elif more_masked == 3:
                    len_y = len(tmp_tf.axe_y)
                    tmp_tf.crop(intervy=[0, len_y - 2], ind=True, inplace=True)
            if not inplace:
                return tmp_tf
        # soft cropping
        else:
            # getting positions to remove
            # (column or line with only masked values)
            axe_y_m = ~np.all(mask_cum, axis=0)
            axe_x_m = ~np.all(mask_cum, axis=1)
            # skip if nothing to do
            if not np.any(axe_y_m) or not np.any(axe_x_m):
                return None
            # getting indices where we need to cut
            axe_x_min = np.where(axe_x_m)[0][0]
            axe_x_max = np.where(axe_x_m)[0][-1]
            axe_y_min = np.where(axe_y_m)[0][0]
            axe_y_max = np.where(axe_y_m)[0][-1]
            # crop
            if inplace:
                self.crop(intervx=[axe_x_min, axe_x_max],
                          intervy=[axe_y_min, axe_y_max], ind=True,
                          inplace=True)
            else:
                tmp_tf = self.copy()
                tmp_tf.crop(intervx=[axe_x_min, axe_x_max],
                            intervy=[axe_y_min, axe_y_max], ind=True,
                            inplace=True)
                return tmp_tf

    def mirroring(self, direction, position, inds_to_mirror='all', mir_coef=1.,
                  inplace=False, interp=None, value=[0, 0]):
        """
        Return the fields with additional mirrored values.

        Parameters
        ----------
        direction : integer
            Axe on which place the symetry plane (1 for x and 2 for y)
        position : number
            Position of the symetry plane along the given axe
        inds_to_mirror : integer
            Number of vector rows to symetrize (default is all)
        mir_coef : number or 2x1 array, optional
            Optional coefficient(s) applied only to the mirrored values.
            It can be an array first value is for 'comp_x' and second one to
            'comp_y' (for vector fields)
        inplace : boolean, optional
            .
        interp : string, optional
            If specified, method used to fill the gap near the
            symetry plane by interpoaltion.
            'value' : fill with the given value,
            'nearest' : fill with the nearest value,
            'linear' (default): fill using linear interpolation
            (Delaunay triangulation),
            'cubic' : fill using cubic interpolation (Delaunay triangulation)
        value : array, optional
            Value at the symetry plane, in case of interpolation
        """
        if inplace:
            tmp_tf = self
        else:
            tmp_tf = self.copy()
        # mirror fields
        for i in range(len(self.fields)):
            tmp_tf.fields[i].mirroring(direction=direction, position=position,
                                       inds_to_mirror=inds_to_mirror,
                                       mir_coef=mir_coef, inplace=True,
                                       interp=interp, value=value)
        # update field
        tmp_tf.__axe_x = self.fields[0].axe_x
        tmp_tf.__axe_y = self.fields[0].axe_y
        # return
        if not inplace:
            return tmp_tf

    def remove_weird_fields(self, std_coef=3.29, treatment='interpolate',
                            inplace=False):
        """
        Look at the time evolution of spatial mean magnitude to identify and
        replace weird fields.

        Parameters
        ----------
        std_coef : number
            Fields associated with mean magnitude outside the interval
            [mean - std_coef*std, mean - std_coef*std] are treated as weird
            fields. Default value of '3.29' corespond for a 99.9% interval.
        treatment : string in ['remove', 'interpolate']
            Type of treatment for the weird fields
            (default is 'interpolate')
        inplace : bool
            .

        Returns
        -------
        tf : TemporalField
            treated temporal field
        """
        # get data
        if inplace:
            tmp_tf = self
        else:
            tmp_tf = self.copy()
        # get weird fields indices
        mean_magn = []
        for field in tmp_tf.fields:
            mean_magn.append(np.sum(field.magnitude[~field.mask]))
        mean = np.mean(mean_magn)
        mean_eps = np.std(mean_magn)*std_coef
        filt = np.logical_or(mean_magn < mean - mean_eps,
                             mean_magn > mean + mean_eps)
        weird_inds = np.arange(len(tmp_tf))[filt]
        # treat weird fields
        if treatment == 'interpolate':
            # replace weird fields with interpolations
            for weird_ind in weird_inds:
                eps = 1
                while True:
                    if (weird_ind + eps in weird_inds or
                            weird_ind - eps in weird_inds):
                        eps += 1
                    else:
                        break
                tmp_tf.fields[weird_ind] = (tmp_tf.fields[weird_ind + eps] +
                                            tmp_tf.fields[weird_ind - eps])/2
        elif treatment == 'remove':
            tmp_tf.remove_fields(weird_inds)
        else:
            raise ValueError()
        # return
        if not inplace:
            return tmp_tf

    def crop(self, intervx=None, intervy=None, intervt=None, full_output=False,
             ind=False, inplace=False):
        """
        Return a croped field in respect with given intervals.

        Parameters
        ----------
        intervx : array, optional
            interval wanted along x
        intervy : array, optional
            interval wanted along y
        intervt : array, optional
            interval wanted along time
        full_output : boolean, optional
            If 'True', cutting indices are alson returned
        inplace : boolean, optional
            If 'True', fields are croped in place.
        """
        # check parameters
        if intervt is not None:
            if not isinstance(intervt, ARRAYTYPES):
                raise TypeError()
            intervt = np.array(intervt, dtype=float)
            if intervt.shape != (2, ):
                raise ValueError()
        # get wanted times
        if intervt is not None:
            if not ind:
                if intervt[0] < self.times[0]:
                    ind1 = 0
                elif intervt[0] > self.times[-1]:
                    raise ValueError()
                else:
                    ind1 = np.where(intervt[0] <= self.times)[0][0]
                if intervt[1] > self.times[-1]:
                    ind2 = len(self.times) - 1
                elif intervt[1] < self.times[0]:
                    raise ValueError()
                else:
                    ind2 = np.where(intervt[1] >= self.times)[0][-1]
                intervt = [ind1, ind2]
            else:
                intervt = [int(intervt[0]), int(intervt[1])]
        # crop
        if inplace:
            cropfield = self
        else:
            cropfield = self.copy()
        # temporal
        if intervt is not None:
            cropfield.fields = cropfield.fields[intervt[0]:intervt[1] + 1]
            cropfield.times = cropfield.times[intervt[0]:intervt[1] + 1]
        # spatial
        fld.Field.crop(cropfield, intervx=intervx, intervy=intervy, ind=ind,
                       inplace=True)
        for field in cropfield.fields:
            field.crop(intervx=intervx, intervy=intervy, ind=ind,
                       inplace=True)
        # returning
        if not inplace:
            return cropfield

    def set_origin(self, x=None, y=None):
        """
        Modify the axis in order to place the origin at the actual point (x, y)

        Parameters
        ----------
        x : number
        y : number
        """
        fld.Field.set_origin(self, x, y)

    def copy(self):
        """
        Return a copy of the velocityfields
        """
        return copy.deepcopy(self)

    def __sort_field_by_time(self):
        if len(self.fields) in [0, 1]:
            return None
        ind_sort = np.argsort(self.times)
        self.times = self.times[ind_sort]
        self.fields = self.fields[ind_sort]

    def display_multiple(self, component=None, kind=None, inds=None,
                         sharecb=False, sharex=False, sharey=False,
                         ncol=None, nrow=None, **plotargs):
        """
        Display a component of the velocity fields.

        Parameters
        ----------
        component : string, optional
            component to display
        kind : string, optional
            Kind of display wanted.
        fields_ind : array of indices
            Indices of fields to display.
        samecb : boolean, optional
            If 'True', the same color system is used for all the fields.
            You have to pass 'vmin' and 'vmax', to have correct results.
        ncol, nrow : int, optional
            Wanted number of columns and rows. If not specified, these values
            are computed so that ncol ~ nrow.
        plotargs : dict, optional
            Arguments passed to the function used to display the vector field.
        """
        nmb_fields = len(inds)
        # getting values
        if component is None or component == 'V':
            try:
                values = [[self.fields[ind].comp_x, self.fields[ind].comp_y]
                          for ind in inds]
            except AttributeError:
                values = [self.fields[ind].values for ind in inds]
        else:
            values = [self.fields[ind].__getattribute__(component)
                      for ind in inds]
        # display
        if sharex or sharey:
            plotargs['adjustable'] = 'datalim'
        db = pplt.Displayer(x=[self.axe_x]*nmb_fields,
                            y=[self.axe_y]*nmb_fields,
                            values=values, kind=kind, **plotargs)
        plot = db.draw_multiple(inds=list(range(len(inds))), sharecb=sharecb,
                                sharex=sharex,
                                sharey=sharey, ncol=ncol, nrow=nrow)
        return plot

    def display(self, compo=None, kind=None, sharecb=True, buffer_size=100,
                **plotargs):
        """
        Create a windows to display temporals field, controlled by buttons.

        Parameters
        ----------
        compo: string
            Component to plot.
        kind: string
            Kind of plot to use.
        sharecb: boolean
            Do all the vector field serie has to share the same colorbar or
            not.
        buffer_size: number
            Number of displays to keep in memory (faster, but use memory).
        **plotargs : dic
            Arguments passed to the plot command.

        Display control
        ---------------
        The display can be controlled useing the button, but also the keyboard:
        space, right arrow or + : next field
        backspace, left arrow or - : previous field
        up arrow : last field
        down arrow : first field
        number + enter : goto a specific frame
        p : play the animated fields
        number + i : set the animation increment
        number + t : set the animation time interval (ms)
        q : close
        s : save an image
        """
        nmb_fields = len(self.fields)
        # getting values
        if compo is None or compo == 'V':
            try:
                values = np.asarray([[field.comp_x, field.comp_y]
                                     for field in self.fields])
            except AttributeError:
                values = np.asarray([field.values for field in self.fields])
        else:
            values = np.asarray([field.__getattribute__(compo)
                                 for field in self.fields])
        # check arguments for colorbar drawning
        if 'norm' in list(plotargs.keys()):
            sharecb = True
            normcb = plotargs['norm']
        else:
            normcb = None
        # display
        db = pplt.Displayer(x=[self.axe_x]*nmb_fields,
                            y=[self.axe_y]*nmb_fields,
                            values=values, kind=kind,
                            buffer_size=buffer_size,
                            **plotargs)
        win = pplt.ButtonManager(db,
                                 xlabel="X " + self.unit_x.strUnit(),
                                 ylabel="Y " + self.unit_y.strUnit(),
                                 sharecb=sharecb, normcb=normcb)
        pplt.DataCursorTextDisplayer(db)
        return db

    def display_animate(self, compo=None, interval=500, fields_inds=None,
                        repeat=True,
                        **plotargs):
        """
        Display fields animated in time.

        Parameters
        ----------
        compo : string
            Composante to display
        interval : number, optionnal
            interval between two frames in milliseconds.
        fields_ind : array of indices
            Indices of wanted fields. by default, all the fields are displayed
        repeat : boolean, optional
            if True, the animation is repeated infinitely.
        additional arguments can be passed (scale, vmin, vmax,...)
        """
        from matplotlib import animation
        if fields_inds is None:
            fields_inds = len(self.fields)
        # getting data
        if self.field_type == vf.VectorField:
            if compo == 'V' or compo is None:
                comp = self.fields
            elif compo == 'magnitude':
                comp = self.magnitude_as_sf
            elif compo == 'x':
                comp = self.Vx_as_sf
            elif compo == 'y':
                comp = self.Vy_as_sf
            elif compo == 'mask':
                comp = self.mask_as_sf
            else:
                raise ValueError()
        elif self.field_type == sf.ScalarField:
            if compo == 'values' or compo is None:
                comp = self.values_as_sf
            elif compo == 'mask':
                comp = self.mask_as_sf
            else:
                raise ValueError()
        else:
            raise TypeError()
        if 'kind' in list(plotargs.keys()):
            kind = plotargs['kind']
        else:
            kind = None
        # display a vector field (quiver)
        if isinstance(comp[0], vf.VectorField)\
                and (kind is None or kind == "quiver"):
            fig = plt.figure()
            ax = plt.gca()
            displ = comp[0].display(**plotargs)
            ttl = plt.title('')
            anim = animation.FuncAnimation(fig, self._update_vf,
                                           frames=fields_inds,
                                           interval=interval, blit=False,
                                           repeat=repeat,
                                           fargs=(fig, ax, displ, ttl, comp,
                                                  compo, plotargs))
            return anim,
        # display a scalar field (contour, contourf or imshow) or a streamplot
        elif isinstance(comp[0], sf.ScalarField)\
                or isinstance(comp[0], vf.VectorField):
            fig = plt.figure()
            ax = plt.gca()
            displ = comp[0].display(**plotargs)
            ttl = plt.suptitle('')
            anim = animation.FuncAnimation(fig, self._update_sf,
                                           frames=fields_inds,
                                           interval=interval, blit=False,
                                           repeat=repeat,
                                           fargs=(fig, ax, displ, ttl, comp,
                                                  compo, plotargs))
            return anim,
        else:
            raise ValueError("I don't know any '{}' composant".format(compo))

    def record_animation(self, anim, filepath, kind='gif', fps=30, dpi=100,
                         bitrate=50, imagemagick_path=None):
        """
        Record an animation in a gif file.
        You must create an animation (using 'display_animate' for example)
        before calling this method.
        You may have to specify the path to imagemagick in orfer to use it.


        Parameters
        ----------

        """
        import matplotlib
        if imagemagick_path is None:
            imagemagick_path = r"C:\Program Files\ImageMagick\convert.exe"
        matplotlib.rc('animation', convert_path=imagemagick_path)
        if kind == 'gif':
            writer = matplotlib.animation.ImageMagickWriter(fps=fps,
                                                            bitrate=bitrate)
            anim.save(filepath, writer=writer, fps=fps, dpi=dpi)
        elif kind == 'mp4':
            anim.save(filepath, writer='fmpeg', fps=fps, bitrate=bitrate)

    def _update_sf(self, num, fig, ax, displ, ttl, comp, compo, plotargs):
        plt.sca(ax)
        ax.cla()
        displ = comp[num]._display(**plotargs)
        title = "{}, at t={:.3} {}"\
            .format(compo, float(self.times[num]),
                    self.unit_times.strUnit())
        ttl.set_text(title)
        return displ,

    def _update_vf(self, num, fig, ax, displ, ttl, comp, compo, plotargs):
        plt.sca(ax)
        ax.cla()
        displ = comp[num]._display(**plotargs)
        title = "{}, at t={:.2f} {}"\
            .format(compo, float(self.times[num]),
                    self.unit_times.strUnit())
        ttl.set_text(title)
        return displ,



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import scipy.interpolate as spinterp
from matplotlib import pyplot as plt

from . import scalarfield as sf, temporalfields as tf
from ..utils.types import ARRAYTYPES, INTEGERTYPES, NUMBERTYPES, STRINGTYPES
from IMTreatment.utils import ProgressCounter
from scipy import ndimage


class TemporalScalarFields(tf.TemporalFields):
    """
    Class representing a set of time-evolving scalar fields.
    """

    @property
    def values_as_sf(self):
        return self

    @property
    def values(self):
        dim = (len(self), self.shape[0], self.shape[1])
        values = np.empty(dim, dtype=float)
        for i, field in enumerate(self.fields):
            values[i, :, :] = field.values[:, :]
        return values

    def get_min_field(self, nmb_min=1):
        """
        Calculate the minimum scalar field, from all the fields.

        Parameters
        ----------
        nmb_min : integer, optional
            Minimum number of values used to take a minimum value.
            Else, the value is masked.
        """
        if len(self.fields) == 0:
            raise ValueError("There is no fields in this object")
        result_f = self.fields[0].copy()
        mask_cum = np.zeros(self.shape, dtype=int)
        mask_cum[np.logical_not(self.fields[0].mask)] += 1
        for field in self.fields[1::]:
            new_min_mask = np.logical_and(field.values < result_f.values,
                                          np.logical_not(field.mask))
            result_f.values[new_min_mask] = field.values[new_min_mask]
            mask_cum[np.logical_not(field.mask)] += 1
        mask = mask_cum <= nmb_min
        result_f.mask = mask
        return result_f

    def get_max_field(self, nmb_min=1):
        """
        Calculate the maximum scalar field, from all the fields.

        Parameters
        ----------
        nmb_min : integer, optional
            Minimum number of values used to take a maximum value.
            Else, the value is masked.
        """
        if len(self.fields) == 0:
            raise ValueError("There is no fields in this object")
        result_f = self.fields[0].copy()
        mask_cum = np.zeros(self.shape, dtype=int)
        mask_cum[np.logical_not(self.fields[0].mask)] += 1
        for field in self.fields[1::]:
            new_max_mask = np.logical_and(field.values > result_f.values,
                                          np.logical_not(field.mask))
            result_f.values[new_max_mask] = field.values[new_max_mask]
            mask_cum[np.logical_not(field.mask)] += 1
        mask = mask_cum <= nmb_min
        result_f.mask = mask
        return result_f

    def get_background(self, noise_level=10, verbose=False):
        """
        Return the background image based on the histogram of values
        at each point.

        Parameters
        ----------
        noise_level: integer
             Level of noise used to differentiate between noise and
             real perturbation of the background image.

        Returns
        -------
        bg: ScalarField object
             Background image
        """
        # Data
        min_len = len(self.times)/10
        values = np.zeros(self.shape)
        if verbose:
            pg = ProgressCounter(init_mess="Computing background image",
                                 nmb_max=self.shape[0]*self.shape[1],
                                 name_things="Pixels",
                                 perc_interv=1)
        # Spatial loop
        for i, j in np.ndindex(self.shape):
            vals = np.array([self.fields[n]._ScalarField__values[i, j]
                             for n in range(len(self.fields))])
            # # HIST
            # hist = np.bincount(vals)
            # val = np.argmax(hist)
            # STATS
            std = np.std(vals, dtype=float)
            old_std = 0
            mean = np.mean(vals, dtype=float)
            while True:
                vals = vals[np.logical_and(vals > mean - std,
                                           vals < mean + std)]
                old_std = std
                std = np.std(vals)
                mean = np.mean(vals)
                if std < noise_level or len(vals) <= min_len or old_std == std:
                    break
            values[i, j] = mean
            if verbose:
                pg.print_progress()
        bg = sf.ScalarField()
        bg.import_from_arrays(axe_x=self.axe_x,
                              axe_y=self.axe_y,
                              values=values,
                              unit_x=self.unit_x,
                              unit_y=self.unit_y)
        return bg

    def substract_background(self, bg, noise_level=10,
                             blend=True,
                             filling_value=None,
                             verbose=False):
        """
        Remove the background without changing the perturbations.

        Parameters
        ----------
        bg: ScalarField object
             Background image
        noise_level: integer
             Level of noise used to differentiate between noise and
             real perturbation of the background image.
        blend: boolean
             Blend the foreground in the background.
        filling_value: integer
             Value used to fill the background.
             Default to the background spatial average.

        Returns
        -------
        nsf: TemporalScalarFields object
            substracted fields.
        """
        # data
        ntf = self.copy()
        bg_mean = bg.mean
        if filling_value is None:
            filling_value = bg_mean
        if verbose:
            pg = ProgressCounter(init_mess="Substracting background image",
                                 nmb_max=len(self.fields),
                                 name_things="Fields")
        # Spatial loop
        for n in range(len(ntf)):
            # Get foreground
            values = ntf.fields[n].values
            mask = np.logical_and(values < bg.values + noise_level,
                                  bg.values - noise_level < values)
            # get foreground and background avg
            if blend:
                fg_mean = np.mean(values[~mask])
                blend_field = (values - bg_mean)/(fg_mean - bg_mean)
                blend_field = ndimage.gaussian_filter(blend_field, 1)
                values = values*blend_field + (1 - blend_field)*filling_value
            values[mask] = filling_value
            ntf.fields[n].values = values
            if verbose:
                pg.print_progress()
        #
        return ntf

    def get_phase_map(self, freq, tf=None, check_spec=None, verbose=True):
        """
        Return the phase map of the temporal scalar field for
        the given frequency.

        Parameters
        ----------
        freq: number
            Wanted frequency
        tf: Integer
            Last time indice to use
        check_spec: Integer
            If not None, specify the number of spectrum to display
            (useful to check if choosen frequencies are relevant).
        verbose: Boolean
            .

        Returns
        -------
        phase_map: ScalarField object
            .
        """
        #
        phases = np.zeros(self.shape, dtype=float)
        norms = np.zeros(self.shape, dtype=float)
        compo = "values"
        if tf is None:
            tf = len(self.fields)
        # select spectrum to display
        if check_spec is not None:
            check_spec_inds = np.random.choice(range(self.shape[0] *
                                                     self.shape[1]),
                                               check_spec)
        # get phases
        if verbose:
            pg = ProgressCounter(init_mess="Computing phase maps",
                                 nmb_max=self.shape[0]*self.shape[1],
                                 name_things="profiles")
        for i, x in enumerate(self.axe_x):
            for j, y in enumerate(self.axe_y):
                if verbose:
                    pg.print_progress()
                # get profile
                profile = self.get_time_profile(compo, (x, y),
                                                wanted_times=[0, tf])
                prof = profile.y
                dx = profile.x[1] - profile.x[0]
                # get fft
                fft = np.fft.fft(prof)
                fft = fft[0:int(len(fft)/2)]
                fft_norm = np.abs(fft)
                fft_phase = np.angle(fft)
                fft_f = np.fft.fftfreq(len(prof), dx)[0:len(fft)]
                # get phase at the wanted frequency
                ind = np.argmin(abs(fft_f - freq))
                # store phases and norms
                phases[i, j] = fft_phase[ind]
                norms[i, j] = fft_norm[ind]
                # display spectrum if asked
                if check_spec:
                    if i*len(self.axe_x)+j in check_spec_inds:
                        fig = plt.figure()
                        ax = fig.add_subplot(1, 1, 1)
                        ax.loglog(fft_f, fft_norm)
                        ax2 = ax.twinx()
                        ax2.plot([], [])
                        ax2.semilogx(fft_f, fft_phase)
                        plt.axvline(fft_f[ind], color="k", ls="--")
                        plt.show(block=True)
        # return
        norm = sf.ScalarField()
        norm.import_from_arrays(axe_x=self.axe_x,
                                axe_y=self.axe_y,
                                values=norms.transpose(),
                                unit_x=self.unit_x,
                                unit_y=self.unit_y,
                                unit_values="")
        phase = sf.ScalarField()
        phase.import_from_arrays(axe_x=self.axe_x,
                                 axe_y=self.axe_y,
                                 values=phases.transpose(),
                                 unit_x=self.unit_x,
                                 unit_y=self.unit_y,
                                 unit_values="rad")
        return norm, phase

    def get_spectral_filtering(self, fmin, fmax, order=2, inplace=False):
        """
        Perform a temporal spectral filtering

        Parameters
        ----------
        fmin, fmax : numbers
            Minimal and maximal frequencies
        order : integer, optional
            Butterworth filter order

        Returns
        -------
        filt_tsf : TemporalScalarFields
            Filtered temporal field
        """
        # prepare
        if inplace:
            tsf = self
        else:
            tsf = self.copy()
        # make spectral filtering on values
        ftsf = self._get_comp_spectral_filtering('values', fmin=fmin,
                                                 fmax=fmax, order=order)
        tsf.fields = ftsf.fields
        # return
        if not inplace:
            return tsf

    def fill(self, tof='spatial', kind='linear', value=0.,
             inplace=False, crop=False):
        """
        Fill the masked part of the array in place.

        Parameters
        ----------
        tof : string
            Can be 'temporal' for temporal interpolation, or 'spatial' for
            spatial interpolation.
        kind : string, optional
            Type of algorithm used to fill.
            'value' : fill with a given value
            'nearest' : fill with nearest available data
            'linear' : fill using linear interpolation
            'cubic' : fill using cubic interpolation
        value : 2x1 array
            Value for filling, '[Vx, Vy]' (only usefull with tof='value')
        inplace : boolean, optional
            .
        crop : boolean, optional
            If 'True', TVF borders are croped before filling.
        """
        # TODO : utiliser Profile.fill au lieu d'une nouvelle méthode de
        #        filling
        # checking parameters coherence
        if len(self.fields) < 3 and tof == 'temporal':
            raise ValueError("Not enough fields to fill with temporal"
                             " interpolation")
        if not isinstance(tof, STRINGTYPES):
            raise TypeError()
        if tof not in ['temporal', 'spatial']:
            raise ValueError()
        if not isinstance(kind, STRINGTYPES):
            raise TypeError()
        if kind not in ['value', 'nearest', 'linear', 'cubic']:
            raise ValueError()
        if crop:
            self.crop_masked_border(hard=False, inplace=True)
        # temporal interpolation
        if tof == 'temporal':
            # getting datas
            # getting super mask (0 where no value are masked and where all
            # values are masked)
            masks = self.mask
            sum_masks = np.sum(masks, axis=0)
            super_mask = np.logical_and(0 < sum_masks,
                                        sum_masks < len(self.fields) - 2)
            # loop on each field position
            for i, j in np.argwhere(super_mask):
                prof = self.get_time_profile('values', i, j, ind=True)
                # creating interpolation function
                if kind == 'value':
                    def interp(x):
                        return value
                elif kind == 'nearest':
                    raise Exception("Not implemented yet")
                elif kind == 'linear':
                    prof_filt = np.logical_not(prof.mask)
                    interp = spinterp.interp1d(prof.x[prof_filt],
                                               prof.y[prof_filt],
                                               kind='linear')

                elif kind == 'cubic':
                    prof_filt = np.logical_not(prof.mask)
                    interp = spinterp.interp1d(prof.x[prof_filt],
                                               prof.y[prof_filt],
                                               kind='cubic')
                else:
                    raise ValueError("Invalid value for 'kind'")
                # inplace or not
                fields = self.fields.copy()
                # loop on all profile masked points
                for ind_masked in prof.mask:
                    try:
                        interp_val = interp(prof.x[prof.mask])
                    except ValueError:
                        continue
                    # putting interpolated value in the field
                    fields[prof.mask].values[i, j] = interp_val
                    fields[prof.mask].mask[i, j] = False
        # spatial interpolation
        elif tof == 'spatial':
            if inplace:
                fields = self.fields
            else:
                tmp_tsf = self.copy()
                fields = tmp_tsf.fields
            for i, field in enumerate(fields):
                fields[i].fill(kind=kind, value=value, inplace=True)
        else:
            raise ValueError("Unknown parameter for 'tof' : {}".format(tof))
        # returning
        if inplace:
            self.fields = fields
        else:
            tmp_tsf = self.copy()
            tmp_tsf.fields = fields
            return tmp_tsf



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from ..utils.types import ARRAYTYPES, INTEGERTYPES, NUMBERTYPES, STRINGTYPES
from ..core import SpatialFields, VectorField, Fields


class SpatialVectorFields(SpatialFields):
    """
    Class representing a set of spatial-evolving velocity fields.
    """

    def __init__(self):
        Fields.__init__(self)
        self.fields_type = VectorField

    @property
    def Vx_as_sf(self):
        return [field.comp_x_as_sf for field in self.fields]

    @property
    def Vy_as_sf(self):
        return [field.comp_y_as_sf for field in self.fields]

    @property
    def magnitude_as_sf(self):
        return [field.magnitude_as_sf for field in self.fields]

    @property
    def theta_as_sf(self):
        return [field.theta_as_sf for field in self.fields]



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import gc
import gzip
import pickle


def memoize_to_file(file_name):
    """ Decorator that memoize the function result in a file. """

    def decorator(original_func):
        gc.disable()
        try:
            with gzip.open(file_name, 'rb') as f:
                cache = pickle.load(f)
        except (IOError, ValueError):
            cache = {}
        gc.enable()

        def new_func(*args, **kwargs):
            key = pickle.dumps((args, kwargs))
            if key not in cache:
                cache[key] = original_func(*args, **kwargs)
                gc.disable()
                with gzip.open(file_name, 'wb') as f:
                    pickle.dump(cache, f, protocol=-1)
                gc.enable()
            return cache[key]

        return new_func

    return decorator



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import psutil
import warnings
import matplotlib.pyplot as plt
from .. import file_operation as imtio
from .. import pod as imtpod
from .. import vortex_detection as imtvod
from os.path import join
import os


class POD_CP_protocol(object):
    def __init__(self, name, imtpath, respath, crop_x=[-np.inf, np.inf],
                 crop_y=[-np.inf, np.inf], crop_t=[-np.inf, np.inf],
                 hard_crop=True,
                 pod_coh=0.05, mirroring=[[2, 0], [1, 0]], eps_traj=15.,
                 red_fact=1., nmb_min_in_traj=1, det_fact=1,
                 thread='all', remove_weird=False):
        self.name = name
        self.imtpath = imtpath
        self.respath = respath
        self.crop_x = crop_x
        self.crop_y = crop_y
        self.crop_t = crop_t
        self.hard_crop = hard_crop
        self.pod = None
        self.pod_coh = pod_coh
        self.mirroring = mirroring
        self.eps_traj = eps_traj
        if det_fact > 1:
            self.detailled = True
        else:
            self.detailled = False
        self.det_fact = det_fact
        self.red_fact = red_fact
        self.nmb_min_in_traj = nmb_min_in_traj
        self.len_data = None
        if thread == "all":
            self.thread = psutil.cpu_coun()
        else:
            self.thread = thread
        self.remove_weird = remove_weird

    def prepare_data(self):
        print("    Preparing data    ")
        # import data
        tvf = imtio.import_from_file(join(self.imtpath,
                                          "{}.cimt".format(self.name)))
        # crop
        if self.hard_crop:
            tvf.crop_masked_border(hard=True, inplace=True)
        else:
            tvf.crop_masked_border(inplace=True)
        tvf.crop(intervx=self.crop_x, intervy=self.crop_y, intervt=self.crop_t,
                 inplace=True)
        # adjust temporal resolution
        if self.red_fact < 1:
            tvf.reduce_temporal_resolution(int(np.round(1./self.red_fact)),
                                           mean=False, inplace=True)
        # check for weird fields
        if self.remove_weird:
            tvf.remove_weird_fields(std_coef=3., treatment='interpolate',
                                    inplace=True)
        # save a display
        mean_field = tvf.get_mean_field()
        mean_field.crop_masked_border(hard=False, inplace=True)
        plt.figure()
        mean_field.display(kind='stream', density=3)
        plt.savefig(join(self.respath, "{}/mean_field.png".format(self.name)))
        plt.close(plt.gcf())
        # store
        imtio.export_to_file(tvf,
                             join(self.imtpath,
                                  "{}_cln.cimt".format(self.name)))
        imtio.export_to_file(mean_field,
                             join(self.imtpath,
                                  "{}_mean.cimt".format(self.name)))
        del tvf

    def pod_decomp(self):
        print("    Making POD decomposition    ")
        # improt data
        tvf = imtio.import_from_file(join(self.imtpath, "{}_cln.cimt"
                                          .format(self.name)))
        # make the decomposition
        pod = imtpod.modal_decomposition(tvf, kind='pod',
                                         wanted_modes='all',
                                         max_vecs_per_node=len(tvf) + 1,
                                         verbose=False)
        # save a display
        pod.display()
        plt.savefig(join(self.respath, "{}/pod.png".format(self.name)))
        plt.close(plt.gcf())
        # save data
        imtio.export_to_file(pod, join(self.imtpath, "{}_pod.cimt"
                                       .format(self.name)))
        del pod

    def pod_reconstr(self, detailled=False):
        if self.detailled:
            print("    Making detailled POD reconstruction    ")
        else:
            print("    Making POD reconstruction    ")
        # improt data
        pod = imtio.import_from_file(join(self.imtpath, "{}_pod.cimt"
                                          .format(self.name)))
        # make reconstruction
        if detailled:
            pod.augment_temporal_resolution(fact=self.det_fact,
                                            interp='linear',
                                            inplace=True)
        wanted_modes = pod.modes_nmb[pod.get_spatial_coherence(raw=True) >
                                     self.pod_coh]
        wanted_modes = np.array(wanted_modes)
        wanted_modes = wanted_modes[wanted_modes < len(pod.modes)/2.]
        rec = pod.reconstruct(wanted_modes=wanted_modes)
        coh = pod.get_spatial_coherence()
        del pod
        # save a display
        plt.figure()
        coh.display(color='k')
        plt.xlim(0, np.max(wanted_modes)*2)
        plt.axhline(self.pod_coh, linestyle='--', color='r')
        plt.plot(wanted_modes, coh.y[wanted_modes], 'or')
        if detailled:
            plt.savefig(join(self.respath, "{}/rec_det.png".format(self.name)))
        else:
            plt.savefig(join(self.respath, "{}/rec.png".format(self.name)))
        plt.close(plt.gcf())
        # save data
        if detailled:
            imtio.export_to_file(rec, join(self.imtpath, "{}_rec_det.cimt"
                                           .format(self.name)))
        else:
            imtio.export_to_file(rec, join(self.imtpath, "{}_rec.cimt"
                                           .format(self.name)))
        del rec

    def CP_detection(self, detailled=False):
        # improt data
        if self.detailled:
            print("    Making detailled CP detection")
        else:
            print("    Making CP detection")
        mem_available = psutil.virtual_memory().free
        if detailled:
            rec = imtio.import_from_file(join(self.imtpath, "{}_rec_det.cimt"
                                              .format(self.name)))
        else:
            rec = imtio.import_from_file(join(self.imtpath, "{}_rec.cimt"
                                              .format(self.name)))
        # Check if there is enough memory to use
        filesize = mem_available - psutil.virtual_memory().free
        possible_nmb_of_copies = int(mem_available/(filesize*2))
        if possible_nmb_of_copies < 1:
            possible_nmb_of_copies = 1
        if possible_nmb_of_copies < self.thread:
            warnings.warn("Use {} instead of {} threads because of "
                          "memory limitation"
                          .format(possible_nmb_of_copies, self.thread))
            self.thread = possible_nmb_of_copies
        # getting cp positions
        traj = imtvod.get_critical_points(rec, kind='pbi',
                                          mirroring=self.mirroring,
                                          verbose=False, thread=self.thread)
        traj.compute_traj(epsilon=self.eps_traj)
        traj.clean_traj(self.nmb_min_in_traj)
        # save display
        plt.figure()
        traj.display_traj('x', filt=[True, True, False, False, False],
                          marker=None)
        if detailled:
            plt.savefig(join(self.respath, "{}/traj_x_cln_det.png"
                             .format(self.name)))
        else:
            plt.savefig(join(self.respath, "{}/traj_x_cln.png"
                             .format(self.name)))
        plt.close(plt.gcf())
        plt.figure()
        traj.display_traj('y', filt=[True, True, False, False, False],
                          marker=None)
        if detailled:
            plt.savefig(join(self.respath, "{}/traj_y_cln_det.png"
                             .format(self.name)))
        else:
            plt.savefig(join(self.respath, "{}/traj_y_cln.png"
                             .format(self.name)))
        plt.close(plt.gcf())
        # save
        if detailled:
            imtio.export_to_file(traj, join(self.imtpath, "{}_traj_det.cimt"
                                            .format(self.name)))
        else:
            imtio.export_to_file(traj, join(self.imtpath, "{}_traj.cimt"
                                            .format(self.name)))
        del traj

    def compute_everything(self):
        if not os.path.exists(join(self.respath, "{}".format(self.name))):
            os.mkdir(join(self.respath, "{}".format(self.name)))
        if not os.path.exists(join(self.imtpath, "{}_cln.cimt"
                                   .format(self.name))):
            self._display_heading()
            self.prepare_data()
            self.pod_decomp()
            self.pod_reconstr()
            self.CP_detection()
        elif not os.path.exists(join(self.imtpath, "{}_pod.cimt"
                                     .format(self.name))):
            self._display_heading()
            self.pod_decomp()
            self.pod_reconstr()
            self.CP_detection()
        elif not os.path.exists(join(self.imtpath, "{}_rec.cimt"
                                     .format(self.name))):
            self._display_heading()
            self.pod_reconstr()
            self.CP_detection()
        elif not os.path.exists(join(self.imtpath, "{}_traj.cimt"
                                     .format(self.name))):
            self._display_heading()
            self.CP_detection()
        else:
            pass
        # detailled study
        if self.detailled:
            if not os.path.exists(join(self.imtpath, "{}_rec_det.cimt"
                                       .format(self.name))):
                self._display_heading()
                self.pod_reconstr(detailled=True)
                self.CP_detection(detailled=True)
            elif not os.path.exists(join(self.imtpath, "{}_traj_det.cimt"
                                         .format(self.name))):
                self._display_heading()
                self.CP_detection(detailled=True)
        else:
            pass

    def recompute_everything(self):
        if not os.path.exists(join(self.respath, "{}".format(self.name))):
            os.mkdir(join(self.respath, "{}".format(self.name)))
        self.prepare_data()
        self.pod_decomp()
        self.pod_reconstr()
        self.CP_detection()

    def _display_heading(self):
        title = "=   {}   =".format(self.name)
        print(("\n" + "="*len(title)))
        print(title)
        print(("="*len(title)))
